# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''键盘输入模块
'''
import time
import ctypes
import win32con
import win32api
import sys
from ctypes import wintypes
import six

_SHIFT = {'~' : '`', '!' : '1', '@' : '2','#' : '3', '$' : '4','%' : '5','^' : '6','&' : '7','*' : '8','(' : '9',')' : '0','_' : '-','+' : '=',
    '{' : '[','}' : ']','|' : '\\',':' : ';','"' : "'",'<' : ',','>' : '.','?' : '/'}

_MODIFIERS = [
    win32con.VK_SHIFT, 
    win32con.VK_CONTROL, 
    win32con.VK_MENU,
    win32con.VK_LWIN,
    win32con.VK_RWIN,
]

_MODIFIER_KEY_MAP = {
    '+': win32con.VK_SHIFT,
    '^': win32con.VK_CONTROL,
    '%': win32con.VK_MENU,                     
}

#ESCAPEDS= '+%^{}'

_CODES = {
    'F1':112, 'F2':113, 'F3':114, 'F4':115, 'F5':116, 'F6':117, 
    'F7':118, 'F8':119, 'F9':120, 'F10':121, 'F11':122, 'F12':123,
    'BKSP':8, 'TAB':9, 'ENTER':13, 'ESC':27, 'END':35, 'HOME':36,'INSERT':45, 'DEL':46, 
    'SPACE':32,'PGUP':33,'PGDN':34,'LEFT':37,'UP':38,'RIGHT':39,'DOWN':40,'PRINT':44,
    'SHIFT':16, 'CTRL':17, 'MENU':18, 'ALT':18,
    'APPS':93, 'CAPS':20, 'WIN':91, 'LWIN': 91, 'RWIN':92, 'NUM':144, 'SCRLK':145
    }

def _scan2vkey(scan):
    return 0xff & ctypes.windll.user32.VkKeyScanW(ctypes.c_wchar("%c" % scan))
    #return 0xff & ctypes.windll.user32.VkKeyScanW(scan)
        
class _KeyboardEvent(object):
        KEYEVENTF_EXTENDEDKEY = 1
        KEYEVENTF_KEYUP  = 2
        KEYEVENTF_UNICODE     = 4
        KEYEVENTF_SCANCODE    = 8

is_64bits = sys.maxsize > 2**32
MAPVK_VK_TO_VSC = 0

if is_64bits:
    wintypes.ULONG_PTR = wintypes.WPARAM
    class _MOUSEINPUT(ctypes.Structure):
        _fields_ = (("dx",          wintypes.LONG),
                    ("dy",         wintypes.LONG),
                    ("mouseData",   wintypes.DWORD),
                    ("dwFlags",     wintypes.DWORD),
                    ("time",        wintypes.DWORD),
                    ("dwExtraInfo", wintypes.ULONG_PTR))

    class _HARDWAREINPUT(ctypes.Structure):
        _fields_ = (("uMsg",    wintypes.DWORD),
                    ("wParamL", ctypes.wintypes.WORD),
                    ("wParamH", ctypes.wintypes.WORD))

    class _KEYBDINPUT(ctypes.Structure):
        _fields_ = (("wVk",         wintypes.WORD),
                    ("wScan",       wintypes.WORD),
                    ("dwFlags",     wintypes.DWORD),
                    ("time",        wintypes.DWORD),
                    ("dwExtraInfo", wintypes.ULONG_PTR))

        def __init__(self, *args, **kwds):
            super(_KEYBDINPUT, self).__init__(*args, **kwds)
            # some programs use the scan code even if KEYEVENTF_SCANCODE
            # isn't set in dwFflags, so attempt to map the correct code.
            if not self.dwFlags & _KeyboardEvent.KEYEVENTF_UNICODE:
                self.wScan = ctypes.windll.user32.MapVirtualKeyExW(self.wVk,
                                                    MAPVK_VK_TO_VSC, 0)

    class _UNION_INPUT_STRUCTS(ctypes.Union):
        "The C Union type representing a single Event of any type"
        _fields_ = [
            ('mi', _MOUSEINPUT),
            ('ki', _KEYBDINPUT),
            ('hi', _HARDWAREINPUT),
        ]

    class _INPUT(ctypes.Structure):
        _anonymous_ = ("_",)
        _fields_ = (("type",   wintypes.DWORD),
                    ("_", _UNION_INPUT_STRUCTS))

    LPINPUT = ctypes.POINTER(_INPUT)
else:
    class _MOUSEINPUT(ctypes.Structure):
        _fields_ = [
            ('dx', ctypes.c_long),
            ('dy', ctypes.c_long),
            ('mouseData', ctypes.c_ulong),
            ('dwFlags', ctypes.c_ulong),
            ('time', ctypes.c_ulong),
            ('dwExtraInfo', ctypes.c_ulong),
        ]

    class _HARDWAREINPUT(ctypes.Structure):
        _fields_ = [
            ('uMsg', ctypes.c_ulong),
            ('wParamL', ctypes.c_ushort),
            ('wParamH', ctypes.c_ushort),
        ]


    class _KEYBDINPUT(ctypes.Structure):
        _fields_ = [
            ('wVk', ctypes.c_ushort),
            ('wScan', ctypes.c_ushort),
            ('dwFlags', ctypes.c_ulong),
            ('time', ctypes.c_ulong),
            ('dwExtraInfo', ctypes.c_ulong),
        ]

    class _UNION_INPUT_STRUCTS(ctypes.Union):
        "The C Union type representing a single Event of any type"
        _fields_ = [
            ('mi', _MOUSEINPUT),
            ('ki', _KEYBDINPUT),
            ('hi', _HARDWAREINPUT),
        ]

    class _INPUT(ctypes.Structure):
        _fields_ = [
            ('type', ctypes.c_ulong),
            ('_', _UNION_INPUT_STRUCTS),
        ]

class KeyInputError(Exception):
    '''键盘输入错误
    '''
    pass

        
class Key(object):
    '''一个按键
    '''

    def __init__(self, key):
        '''Constructor
        
        :type key: number or charactor
        :param key: 按键 
        '''
        self._flag = 0
        self._modifiers = []
        if isinstance(key, six.string_types):
            self._scan = ord(key)
            if self._scan < 256: #ASCII code
                self._vk = _scan2vkey(self._scan)
                if key.isupper() or key in _SHIFT: #按下shift键
                    self._modifiers.append(Key(_MODIFIER_KEY_MAP['+']))
            else: #unicode
                self._vk = 0
                self._flag |= _KeyboardEvent.KEYEVENTF_UNICODE
        elif isinstance(key, int): #virtual Key
            self._vk = key
            self._scan = ctypes.windll.user32.MapVirtualKeyW(self._vk, 0)
            if self._isExtendedKey(self._vk):
                self._flag |= _KeyboardEvent.KEYEVENTF_EXTENDEDKEY
        else:
            raise KeyInputError('Key is not a number or string')
        
        # 若是以下虚拟键则需要同步
        # 暂只对最主要的SHIFT、CTRL、ALT键进行同步
        self._SyncVKeys = [win32con.VK_SHIFT, win32con.VK_LSHIFT, win32con.VK_RSHIFT,
         win32con.VK_CONTROL, win32con.VK_LCONTROL, win32con.VK_RCONTROL,
         win32con.VK_MENU, win32con.VK_LMENU, win32con.VK_RMENU]
    
    def appendModifierKey(self, key):
        '''Modifier Key comes with the key
        
        :type key: Key
        :param key: Ctrl, Shift or Atl Key 
        '''
        self._modifiers.append(key)

    def _isExtendedKey(self, vkey):
        if ((vkey >= 33 and vkey <= 46) or 
            (vkey >= 91 and vkey <= 93) ):
            return True
        else:
            return False
        
    def _inputKey(self, up):
        inp = _INPUT()
        inp.type = 1
        inp._.ki.wVk = self._vk
        inp._.ki.wScan = self._scan
        inp._.ki.dwFlags |= self._flag
        if up:
            inp._.ki.dwFlags |= _KeyboardEvent.KEYEVENTF_KEYUP
        ctypes.windll.user32.SendInput(1,ctypes.byref(inp),ctypes.sizeof(_INPUT))
    
    def inputKey(self):
        '''键盘模拟输入按键
        '''
        for mkey in self._modifiers:
            mkey._inputKey(up=False)
        
        self._inputKey(up=False)
        self._inputKey(up=True)
        for mkey in self._modifiers:
            mkey._inputKey(up=True)
            
    def _postKey(self, hwnd, up):
        '''给某个窗口发送按钮
        '''
        if up:
            ctypes.windll.user32.PostMessageW(hwnd, win32con.WM_KEYUP, self._vk, self._scan <<16 | 0xc0000001)
        else:
            ctypes.windll.user32.PostMessageW(hwnd, win32con.WM_KEYDOWN, self._vk, self._scan<<16 | 1)
        time.sleep(0.01) #必须加，否则有些控件响应不了，会产生问题
        
    def postKey(self, hwnd):
        '''将按键消息发到hwnd
        '''
        for mkey in self._modifiers:
            mkey._inputKey(up=False)
            time.sleep(0.01) #必须加，否则下面实际寄送的键盘按键消息可能比这个消息更快到达目标窗口

        if self._scan < 256:
            self._postKey(hwnd, up=False)
            self._postKey(hwnd, up=True)
        else:
            ctypes.windll.user32.PostMessageW(hwnd, win32con.WM_CHAR, self._scan, 1)
        
        for mkey in self._modifiers:
            mkey._inputKey(up=True)
    
    def _isPressed(self):
        """该键是否被按下
        """
        if (win32api.GetAsyncKeyState(self._vk) & 0x8000) == 0:
            return False
        else:
            return True
        
    def _isToggled(self):
        """该键是否被开启，如Caps Lock或Num Lock等
        """
        if(win32api.GetKeyState(self._vk) & 1):
            return True
        else:
            return False
    
class Keyboard(object):
    '''键盘输入类，实现了两种键盘输入方式。
    
    一类方法使用模拟键盘输入的方式。
    另一类方法使用Windows消息的机制将字符串直接发送的窗口。
    
    键盘输入类支持以下字符的输入。
    1、特殊字符：^, +, %,  {, }
        '^'表示Control键，同'{CTRL}'。'+'表示Shift键，同'{SHIFT}'。'%'表示Alt键，同'{ALT}'。
        '^', '+', '%'可以单独或同时使用，如'^a'表示CTRL+a，’^%a'表示CTRL+ALT+a。
        {}： 大括号用来输入特殊字符本身和虚键，如‘{+}’输入加号，'{F1}'输入F1虚键，'{}}'表示输入'}'字符。 
    2、ASCII字符：除了特殊字符需要｛｝来转义，其他ASCII码字符直接输入，
    3、Unicode字符：直接输入，如"测试"。
    4、虚键：
        {F1}, {F2},...{F12}
        {Tab},{CAPS},{ESC},{BKSP},{HOME},{INSERT},{DEL},{END},{ENTER}
        {PGUP},{PGDN},{LEFT},{RIGHT},{UP},{DOWN},{CTRL},{SHIFT},{ALT},{APPS}..
           注意：当使用联合键时，注意此类的问题,inputKeys('^W')和inputKeys('%w')，字母'w'的大小写产生的效果可能不一样
    '''
    _keyclass = Key
    _pressedkey = None
    
    @staticmethod
    def selectKeyClass(newkeyclass):
        oldkeyclass = Keyboard._keyclass
        Keyboard._keyclass = newkeyclass
        return oldkeyclass
    
    @staticmethod
    def _parse_keys(keystring):        
        keys = []
        modifiers = []
        index = 0
        while index < len(keystring):
            c = keystring[index]
            index += 1
                
            # Escape or named key
            if c == "{":
                end_pos = keystring.find("}", index)
                if end_pos == -1:
                    raise KeyInputError('`}` not found')
                elif end_pos == index and keystring[end_pos+1] == '}': #{}}
                        index +=2
                        code = '}'
                else: 
                    code = keystring[index:end_pos]
                    index = end_pos + 1
                    
                if code in _CODES.keys():
                    key = _CODES[code]
                elif len(code) == 1:
                    key = code
                else:
                    raise KeyInputError("Code '%s' is not supported" % code)
            
            # unmatched "}"
            elif c == '}':
                raise KeyInputError('`}` should be preceeded by `{`')
            
            elif c in _MODIFIER_KEY_MAP.keys():
                key = _MODIFIER_KEY_MAP[c]
       
            # so it is a normal character
            else:
                key = c
            if key in _MODIFIERS :
                modifiers.append(key)
            else:
                akey = Keyboard._keyclass(key)
                for mkey in modifiers:
                    akey.appendModifierKey(Keyboard._keyclass(mkey))
                modifiers = []
                keys.append(akey)
        for akey in modifiers:
            keys.append(Keyboard._keyclass(akey))
            
        return keys
    
    @staticmethod
    def inputKeys(keys, interval=0.01):
        '''模拟键盘输入字符串
        
        :type keys: utf-8 str or unicode
        :param keys: 键盘输入字符串,可输入组合键，如"{CTRL}{MENU}a"
        :type interval: number
        :param interval: 输入的字符和字符之间的暂停间隔。
        '''

        if not isinstance(keys, six.text_type):
            keys = keys.decode('utf-8')
        keys = Keyboard._parse_keys(keys)
    
        for k in keys:
            k.inputKey()
            time.sleep(interval)
            
   
    @staticmethod
    def postKeys(hwnd, keys, interval=0.01):
        '''将字符串以窗口消息的方式发送到指定win32窗口。
        
        :type hwnd: number
        :param hwnd: windows窗口句柄 
        :type keys: utf8 str 或者 unicode
        :param keys: 键盘输入字符串
        :type interval: number
        :param interval: 输入的字符和字符之间的暂停间隔。
        '''
        if not isinstance(keys, six.text_type):
            keys = keys.decode('utf-8')
        keys = Keyboard._parse_keys(keys)
    
        for k in keys:
            k.postKey(hwnd)
            time.sleep(interval)
            
    @staticmethod
    def pressKey(key):
        """按下某个键
        """
        if Keyboard._pressedkey:
            raise ValueError("尚有按键未释放,请先对按键进行释放,未释放的按键为: %s"%Keyboard._pressedkey)
            
        if not isinstance(key, six.text_type):
            key = key.decode('utf-8')
        keys = Keyboard._parse_keys(key)
        if len(keys) != 1:
            raise ValueError("输入参数错误,只支持输入一个键,key: %s"%key)  
        keys[0]._inputKey(up=False)
        Keyboard._pressedkey = key
                
    @staticmethod
    def releaseKey(key=None):
        """释放上一个被按下的键
        """
        #后续改成不带参数的，将释放上一个按键，暂时兼容下现有的使用方式
        if not Keyboard._pressedkey:
            raise Exception("没有可释放的按键")
        
        key = Keyboard._pressedkey                
        if not isinstance(key, six.text_type):
            key = key.decode('utf-8')
        keys = Keyboard._parse_keys(key)
        if len(keys) != 1:
            raise ValueError("输入参数错误,只支持输入一个键,key: %s"%key)  
        keys[0]._inputKey(up=True)      
        Keyboard._pressedkey = None  
        
    @staticmethod
    def isPressed(key):
        """是否被按下
        """
        if not isinstance(key, six.text_type):
            key = key.decode('utf-8')
        keys = Keyboard._parse_keys(key)
        if len(keys) != 1:
            raise ValueError("输入参数错误,只支持输入一个键,key: %s"%key)  
        return keys[0]._isPressed()
    
    @staticmethod
    def clear():
        """释放被按下的按键
        """
        if Keyboard._pressedkey:
            Keyboard.releaseKey()

    @staticmethod
    def isTroggled(key):
        """是否开启，如Caps Lock或Num Lock等
        """
        if not isinstance(key, six.text_type):
            key = key.decode('utf-8')
        keys = Keyboard._parse_keys(key)
        if len(keys) != 1:
            raise ValueError("输入参数错误,只支持输入一个键,key: %s"%key)  
        return keys[0]._isToggled()
    
if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
"""
鼠标操作模块
"""

import win32api
import win32con
import win32gui
import winerror
import time

class MouseFlag(object):
    '''鼠标按键枚举类
    '''
    LeftButton, MiddleButton, RightButton=range(3)

class MouseClickType(object):
    '''鼠标点击枚举类
    '''
    SingleClick, DoubleClick = range(2)    

class MouseCursorType(object):
    '''鼠标图标枚举类型
    '''
    Arrow, Hand, Busy, Cross, No, Help, IBeam = range(7) 
    
_cursor_flags = { 
                 0:win32con.IDC_ARROW,
                 1:win32con.IDC_HAND,
                 2:win32con.IDC_WAIT,
                 3:win32con.IDC_CROSS,
                 4:win32con.IDC_NO,
                 5:win32con.IDC_HELP,
                 6:win32con.IDC_IBEAM,
                 }
    
_mouse_msg={
    MouseFlag.LeftButton: [win32con.MOUSEEVENTF_LEFTDOWN, win32con.MOUSEEVENTF_LEFTUP],
    MouseFlag.MiddleButton: [win32con.MOUSEEVENTF_MIDDLEDOWN, win32con.MOUSEEVENTF_MIDDLEUP],
    MouseFlag.RightButton: [win32con.MOUSEEVENTF_RIGHTDOWN, win32con.MOUSEEVENTF_RIGHTUP]
    }

_mouse_msg_param={
    MouseFlag.LeftButton: [win32con.WM_LBUTTONDOWN, win32con.WM_LBUTTONUP, 1, win32con.WM_LBUTTONDBLCLK], 
    MouseFlag.MiddleButton:  [win32con.WM_MBUTTONDOWN, win32con.WM_MBUTTONUP, 16, win32con.WM_RBUTTONDBLCLK],
    MouseFlag.RightButton: [win32con.WM_RBUTTONDOWN, win32con.WM_RBUTTONUP, 2, win32con.WM_MBUTTONDBLCLK]
    }

_mouse_ncmsg_param={
    MouseFlag.LeftButton: [win32con.WM_NCLBUTTONDOWN, win32con.WM_LBUTTONUP, 1, win32con.WM_NCLBUTTONDBLCLK], 
    }


class Mouse(object):
    '''鼠标操作静态类
    '''
    
    _last_click_time = time.time()
    
    @staticmethod
    def click(x, y, flag=MouseFlag.LeftButton, 
              clicktype=MouseClickType.SingleClick):
        """鼠标点击(x,y)点
        
        :param x: 屏幕x坐标
        :type x: int  
        :param y: 屏幕y坐标
        :type y: int
        :param mouseFlag: 鼠标按钮
        :type mouseFlag: qt4c.mouse.MouseFlag
        :param clickType: 鼠标动作,如双击还是单击
        :type clickType: qt4c.mouse.MouseClickType
        """
        win32api.SetCursorPos((x,y))
        win32api.mouse_event(_mouse_msg[flag][0] | _mouse_msg[flag][1],0,0,0,0)
        if clicktype != MouseClickType.SingleClick:
            time.sleep(0.1)
            win32api.mouse_event(_mouse_msg[flag][0] | _mouse_msg[flag][1],0,0,0,0)
    
    @staticmethod
    def _clickSlowly(x,y,flag=MouseFlag.LeftButton, interval=0.1):
        """模拟鼠标缓慢点击，在鼠标键按下和释放之间存在一个interval的时间间隔
        """
        win32api.SetCursorPos((x,y))
        win32api.mouse_event(_mouse_msg[flag][0],0,0,0,0)
        if interval:
            time.sleep(interval)
        win32api.mouse_event(_mouse_msg[flag][1],0,0,0,0)
        
    @staticmethod
    def sendClick(hwnd, x, y, flag=MouseFlag.LeftButton, 
                  clicktype=MouseClickType.SingleClick):
        """在目标窗口通过SendMessage方式产生鼠标点击的动作
        
        :param hwnd: 目标窗口句柄
        :type hwnd: 整数
        :param x: 屏幕x坐标
        :type x: 整数  
        :param y: 屏幕y坐标
        :type y: 整数
        :param flag: 鼠标键类型 
        :type flag: 枚举类型, MouseFlag.LeftButton|MouseFlag.MiddleButton|MouseFlag.RightButton
        :param clicktype: 鼠标键点击方式
        :type clicktype: 枚举类型, MouseClickType.SingleClick | MouseClickType.DoubleClick
        """
        
#        sleep_time = (win32gui.GetDoubleClickTime() / 1000.0 + 0.05) -  (time.time() - Mouse._last_click_time)
#        if sleep_time > 0 :
#            time.sleep(sleep_time)
        keystate = 0
        if win32api.GetAsyncKeyState(win32con.VK_CONTROL) & 0x8000 != 0:
            keystate = keystate | win32con.MK_CONTROL
        if win32api.GetAsyncKeyState(win32con.VK_SHIFT) & 0x8000 != 0:
            keystate = keystate | win32con.MK_SHIFT
        win32api.SetCursorPos((x,y))

        time.sleep(0.2)
        c_x, c_y = win32gui.ScreenToClient(hwnd, (x, y))
        if clicktype == MouseClickType.SingleClick:
            try:
                win32gui.SendMessageTimeout(hwnd, _mouse_msg_param[flag][0], _mouse_msg_param[flag][2] | keystate, (c_y << 16) | c_x, 0, 1000)
            except win32gui.error as e:
                if e.winerror == winerror.ERROR_INVALID_WINDOW_HANDLE: #无效窗口句柄，有些窗口post第一个消息后就会消失
                    pass
                elif e.winerror == winerror.ERROR_TIMEOUT: #timeout
                    win32gui.SendMessageTimeout(hwnd, win32con.WM_NULL, 0, 0, 0, 30000) #发送NULL确认上个消息已处理
                else:
                    raise e
            try:
                win32gui.SendMessageTimeout(hwnd, _mouse_msg_param[flag][1], 0 | keystate, (c_y << 16) | (c_x) , 0, 1000)
            except win32gui.error as e:
                if e.winerror == winerror.ERROR_INVALID_WINDOW_HANDLE: #无效窗口句柄，有些窗口post第一个消息后就会消失
                    pass
                elif e.winerror == winerror.ERROR_TIMEOUT:#timeout
                    win32gui.SendMessageTimeout(hwnd, win32con.WM_NULL, 0, 0, 0, 30000)#发送NULL确认上个消息已处理
                else:
                    raise e 
        else:
            win32gui.SendMessage(hwnd, _mouse_msg_param[flag][0], _mouse_msg_param[flag][2] | keystate, (c_y << 16) | c_x)
            win32gui.SendMessage(hwnd, _mouse_msg_param[flag][1], 0 | keystate, (c_y << 16) | (c_x) )
#            time.sleep(0.1)
            win32gui.SendMessage(hwnd, _mouse_msg_param[flag][3], _mouse_msg_param[flag][2] | keystate, (c_y << 16) | c_x)
            win32gui.SendMessage(hwnd, _mouse_msg_param[flag][1], 0 | keystate, (c_y << 16) | (c_x) )
#        Mouse._last_click_time = time.time()
        
    @staticmethod
    def postClick(hwnd, x, y, flag=MouseFlag.LeftButton, 
                  clicktype=MouseClickType.SingleClick):
        """在目标窗口通过PostMessage的方式产生鼠标点击的动作
        
        :param hwnd: 目标窗口句柄
        :type hwnd: 整数
        :param x: 屏幕x坐标
        :type x: 整数  
        :param y: 屏幕y坐标
        :type y: 整数
        :param flag: 鼠标键类型 
        :type flag: 枚举类型, MouseFlag.LeftButton|MouseFlag.MiddleButton|MouseFlag.RightButton
        :param clicktype: 鼠标键点击方式
        :type clicktype: 枚举类型, MouseClickType.SingleClick | MouseClickType.DoubleClick
        """
        sleep_time = (win32gui.GetDoubleClickTime() / 1000.0 + 0.05) -  (time.time() - Mouse._last_click_time)
        if sleep_time > 0 :
            time.sleep(sleep_time)
        keystate = 0
        if win32api.GetAsyncKeyState(win32con.VK_CONTROL) & 0x8000 != 0:
            keystate = keystate | win32con.MK_CONTROL
        if win32api.GetAsyncKeyState(win32con.VK_SHIFT) & 0x8000 != 0:
            keystate = keystate | win32con.MK_SHIFT
        win32api.SetCursorPos((x,y))
        c_x, c_y = win32gui.ScreenToClient(hwnd, (x, y))
        if clicktype == MouseClickType.SingleClick:
            try:
                win32gui.PostMessage(hwnd, win32con.WM_NULL, 0 , 0)
                win32gui.PostMessage(hwnd, _mouse_msg_param[flag][0], _mouse_msg_param[flag][2] | keystate, (c_y << 16) | c_x)
                win32gui.PostMessage(hwnd, _mouse_msg_param[flag][1], 0 | keystate, (c_y << 16) | (c_x) )
            except win32gui.error as e:
                if e.winerror == winerror.ERROR_INVALID_WINDOW_HANDLE: #无效窗口句柄，有些窗口post第一个消息后就会消失
                    pass
                else:
                    raise e
        else:
            win32gui.PostMessage(hwnd, _mouse_msg_param[flag][0], _mouse_msg_param[flag][2] | keystate, (c_y << 16) | c_x)
            win32gui.PostMessage(hwnd, _mouse_msg_param[flag][1], 0 | keystate, (c_y << 16) | (c_x) )
            time.sleep(0.1)
            win32gui.PostMessage(hwnd, _mouse_msg_param[flag][3], _mouse_msg_param[flag][2] | keystate, (c_y << 16) | c_x)
            win32gui.PostMessage(hwnd, _mouse_msg_param[flag][1], 0 | keystate, (c_y << 16) | (c_x) )
        Mouse._last_click_time = time.time()
        
        
#    @staticmethod
#    def _sendClickSlowly(hwnd, x, y, flag=MouseFlag.LeftButton, interval=0.1):
#        """在目标窗口通过发消息的方式产生鼠标点击的动作，
#                             但在鼠标键按下和释放之间存在一个interval的时间间隔
#        """
#        win32api.SetCursorPos((x,y))
#        c_x, c_y = win32gui.ScreenToClient(hwnd, (x, y))
#        win32gui.SendMessage(hwnd, win32con.WM_NCHITTEST, 0, (c_y << 16) | c_x)
#        try:
#            win32gui.PostMessage(hwnd, _mouse_msg_param[flag][0], _mouse_msg_param[flag][2], (c_y << 16) | c_x)
#            if interval:
#                time.sleep(interval)
#            win32gui.PostMessage(hwnd, _mouse_msg_param[flag][1], 0, (c_y << 16) | (c_x) )
#        except win32gui.error as e:
#            if e[0] == 1400: #无效窗口句柄，有些窗口post第一个消息后就会消失
#                pass
#            else:
#                raise e
            
    @staticmethod
    def sendNCClick(hwnd, x, y, flag=MouseFlag.LeftButton, 
                  clicktype=MouseClickType.SingleClick):
        """在目标窗口的Non-Client区域通过发消息的方式产生鼠标点击的动作
        
        :param hwnd: 目标窗口句柄
        :type hwnd: 整数
        :param x: 屏幕x坐标
        :type x: 整数  
        :param y: 屏幕y坐标
        :type y: 整数
        :param flag: 鼠标键类型 
        :type flag: 枚举类型, MouseFlag.LeftButton|MouseFlag.MiddleButton|MouseFlag.RightButton
        :param clicktype: 鼠标键点击方式
        :type clicktype: 枚举类型, MouseClickType.SingleClick | MouseClickType.DoubleClick
        """
        win32api.SetCursorPos((x,y))
        c_x, c_y = win32gui.ScreenToClient(hwnd, (x, y))
        uHitTestResult = win32gui.SendMessage(hwnd, win32con.WM_NCHITTEST, 0, (c_y << 16) | c_x)
        if clicktype == MouseClickType.SingleClick:
            win32gui.PostMessage(hwnd, _mouse_ncmsg_param[flag][0], uHitTestResult, (c_y << 16) | c_x)
            win32gui.PostMessage(hwnd, _mouse_msg_param[flag][1], 0, (c_y << 16) | (c_x) )
        else:
            win32gui.PostMessage(hwnd, _mouse_ncmsg_param[flag][0], uHitTestResult, (c_y << 16) | c_x)
            win32gui.PostMessage(hwnd, _mouse_ncmsg_param[flag][1], 0, (c_y << 16) | (c_x) )
            time.sleep(0.1)
            win32gui.PostMessage(hwnd, _mouse_ncmsg_param[flag][3], uHitTestResult, (c_y << 16) | c_x)
            win32gui.PostMessage(hwnd, _mouse_ncmsg_param[flag][1], 0, (c_y << 16) | (c_x) )
            
    @staticmethod
    def drag(fromX, fromY, toX, toY, flag=MouseFlag.LeftButton, intervaltime=1):
        """鼠标从(fromX, fromX)拖拽到(toX, toY)
        
        :param fromX: 屏幕x坐标
        :type fromX: 整数  
        :param fromY: 屏幕y坐标
        :type fromY: 整数
        :param toX: 屏幕x坐标
        :type toX: 整数  
        :param toY: 屏幕y坐标
        :type toY: 整数
        :param flag: 鼠标键类型 
        :type flag: 枚举类型, MouseFlag.LeftButton|MouseFlag.MiddleButton|MouseFlag.RightButton
        """
        win32api.SetCursorPos((fromX,fromY))
        time.sleep(0.1)
        win32api.mouse_event(_mouse_msg[flag][0],0,0,0,0)
        time.sleep(intervaltime)
        win32api.SetCursorPos((toX,toY))
        time.sleep(0.1)
        win32api.mouse_event(_mouse_msg[flag][1],0,0,0,0)
    
    @staticmethod
    def press(x, y, flag=MouseFlag.LeftButton):
        """在某个位置按下鼠标键
        
        :param x: 屏幕x坐标
        :type x: 整数  
        :param y: 屏幕y坐标
        :type y: 整数
        :param flag: 鼠标键类型 
        :type flag: 枚举类型, MouseFlag.LeftButton|MouseFlag.MiddleButton|MouseFlag.RightButton
        """ 
        win32api.SetCursorPos((x,y))
        time.sleep(0.1)
        win32api.mouse_event(_mouse_msg[flag][0],0,0,0,0)
        
    @staticmethod
    def release(x, y, flag=MouseFlag.LeftButton):
        """在某个位置释放鼠标键，与press配对使用
        
        :param x: 屏幕x坐标
        :type x: 整数  
        :param y: 屏幕y坐标
        :type y: 整数
        :param flag: 鼠标键类型 
        :type flag: 枚举类型, MouseFlag.LeftButton|MouseFlag.MiddleButton|MouseFlag.RightButton
        """ 
        win32api.SetCursorPos((x,y))
        time.sleep(0.1)
        win32api.mouse_event(_mouse_msg[flag][1],0,0,0,0)
        
    @staticmethod
    def postMove(hwnd, toX, toY):
        win32api.SetCursorPos((toX,toY))
        c_x, c_y = win32gui.ScreenToClient(hwnd, (toX, toY))
        keystate = 0
        if win32api.GetAsyncKeyState(win32con.VK_CONTROL) & 0x8000 != 0:
            keystate = keystate | win32con.MK_CONTROL
        if win32api.GetAsyncKeyState(win32con.VK_SHIFT) & 0x8000 != 0:
            keystate = keystate | win32con.MK_SHIFT
        win32gui.PostMessage(hwnd, win32con.WM_MOUSEMOVE, keystate, (c_y << 16) | c_x)
        
    @staticmethod
    def sendMove(hwnd, toX, toY):
        win32api.SetCursorPos((toX,toY))
        c_x, c_y = win32gui.ScreenToClient(hwnd, (toX, toY))
        keystate = 0
        if win32api.GetAsyncKeyState(win32con.VK_CONTROL) & 0x8000 != 0:
            keystate = keystate | win32con.MK_CONTROL
        if win32api.GetAsyncKeyState(win32con.VK_SHIFT) & 0x8000 != 0:
            keystate = keystate | win32con.MK_SHIFT
        win32gui.SendMessage(hwnd, win32con.WM_MOUSEMOVE, keystate, (c_y << 16) | c_x)
    
    @staticmethod
    def move(toX, toY):
        """鼠标移动到(tox,toy)
        
        :param x: 屏幕x坐标
        :type x: int  
        :param y: 屏幕y坐标
        :type y: int
        """
        win32api.SetCursorPos((toX,toY))
    
    @staticmethod
    def getPosition():
        '''当前Mouse的位置
        '''
        x,y = win32api.GetCursorPos()
        return (x,y)
        
    @staticmethod
    def getCursorType():
        """返回当前鼠标图标类型
        
        :rtype: MouseCursorType
        """
        hcursor = win32gui.GetCursorInfo()[1]
        for key, flag in _cursor_flags.items():
            if win32gui.LoadCursor(0, flag) == hcursor:
                return key
        return None 
                    
    @staticmethod
    def scroll(bForward=False):
        '''鼠标滚动
        bForward: True则向前滚动，False则向后滚动。默认是False。
        '''
        if(bForward==True):
            win32api.mouse_event(win32con.MOUSEEVENTF_WHEEL,0,0,120)
        else:
            win32api.mouse_event(win32con.MOUSEEVENTF_WHEEL,0,0,-120)

                                                              
if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''
TUIA测试用例基类
'''
from __future__ import division
import time
import threading
import pythoncom
import six
from testbase.testcase import TestCase
from qt4c import util

def _screenshot(path):
    '''save screenshot
    '''
    
    import win32gui, win32con
    import qt4c.wintypes as wintypes
    import ctypes

    hdcScreen = ctypes.windll.gdi32.CreateDCA(six.b('DISPLAY'), None, None, None)
    hdcCapture = ctypes.windll.gdi32.CreateCompatibleDC(hdcScreen)
    
    hwnd_dsk = win32gui.GetDesktopWindow()
    l, t, r, b = win32gui.GetWindowRect(hwnd_dsk)
    w = r - l
    h = b - t
    ratio = util.getDpi()
    w = int(w * ratio)
    h = int(h * ratio)
    
    lpBits = ctypes.c_void_p(0)
    bmiCapture = wintypes.BITMAPINFO()
    bmiCapture.bmiHeader.biSize = ctypes.sizeof(wintypes.BITMAPINFO)
    bmiCapture.bmiHeader.biWidth = w
    bmiCapture.bmiHeader.biHeight = h
    bmiCapture.bmiHeader.biPlanes = 1
    bmiCapture.bmiHeader.biBitCount = 24
    bmiCapture.bmiHeader.biCompression = win32con.BI_RGB
    bmiCapture.bmiHeader.biSizeImage = 0
    bmiCapture.bmiHeader.biXPelsPerMeter = 0
    bmiCapture.bmiHeader.biYPelsPerMeter = 0 
    bmiCapture.bmiHeader.biClrUsed = 0
    bmiCapture.bmiHeader.biClrImportant = 0
    
    hbmCapture = ctypes.windll.gdi32.CreateDIBSection(None,
                                                         ctypes.byref(bmiCapture),
                                                         win32con.DIB_RGB_COLORS,
                                                         ctypes.byref(lpBits),
                                                         None,
                                                         0)

    if hbmCapture:
        CAPTUREBLT = 0x40000000
        hbmOld = ctypes.windll.gdi32.SelectObject(hdcCapture, hbmCapture)
        ctypes.windll.gdi32.BitBlt(hdcCapture,
                                   0,
                                   0,
                                   w,
                                   h,
                                   hdcScreen,
                                   0,
                                   0,
                                   win32con.SRCCOPY | CAPTUREBLT)
        flag, hcursor, (xcur, ycur) = win32gui.GetCursorInfo()
        if flag == win32con.CURSOR_SHOWING:
            iconinfo = win32gui.GetIconInfo(hcursor)
            ctypes.windll.user32.DrawIcon(hdcCapture,
                                          int(xcur * ratio - iconinfo[1] * ratio),
                                          int(ycur * ratio - iconinfo[2] * ratio),
                                          hcursor)
        ctypes.windll.gdi32.SelectObject(hdcCapture, hbmOld)
    
    ctypes.windll.gdi32.DeleteDC(hdcCapture)
    ctypes.windll.gdi32.DeleteDC(hdcScreen)
    
    #位图像素在位元以行为单位对齐存储，每一行的总像素大小都向上取证为4字节
    #参考http://en.wikipedia.org/wiki/BMP_file_format
    alloced_bytes = (((w * 3) % 4) and ((w * 3) + (w * 3) % 4) or w * 3) * h
    pBuf = ctypes.create_string_buffer(alloced_bytes)
    cpy_bytes = ctypes.windll.gdi32.GetBitmapBits(hbmCapture,
                                                  alloced_bytes,
                                                  ctypes.byref(pBuf))
    ctypes.windll.gdi32.DeleteObject(hbmCapture)
    
    if cpy_bytes == 0:
        raise RuntimeError("failed to call GetBitmapBits!")
    
    #保留Image模块，方便使用其保存成文件的功能
    try:
        from PIL import Image
    except ImportError: #PIL模块Import失败时尝试Import Pillow
        import Image
    
    im = Image.frombuffer('RGB', (w, h), pBuf, 'raw', 'BGR', 0, 1)
    im.save(path)
    

class ClientTestCase(TestCase):
    '''QT4C测试用例基类
    '''
    
    def initTest(self, testresult):
        '''测试用例初始化。慎用此函数，尽量将初始化放到preTest里。
        '''
        super(ClientTestCase, self).initTest(testresult)
        self.__com_threadid = threading.current_thread().ident
        pythoncom.CoInitialize()  # @UndefinedVariable

    def cleanTest(self):
        '''测试用例反初始化
        '''
        if self.__com_threadid == threading.current_thread().ident:
            pythoncom.CoUninitialize()  # @UndefinedVariable
        #清空测试桩对象缓存（这里有个隐含bug,早前的线程创建的对象有可能会被后来线程的使用，因为有可能生成了相同的线程id
        super(ClientTestCase, self).cleanTest()
        
    def get_extra_fail_record(self):
        timestr = time.strftime('%y%m%d%H%M%S', time.localtime())
        file_path = "%s_%s.jpg" % (self.TestClassName, timestr)
        _screenshot(file_path)
        return {}, {'截图': file_path}

    init_test = initTest
    clean_test = cleanTest

if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''IE驱动模块

IE中的坑：
1、为避免用户传入的js存在语法错误，使用eval方式执行；这种方式可以获得最后一句话的返回值
2、eval中使用var xxx=123;不能定义变量，需要使用window['xxx'] = 123; 改成使用window.eval可以解决，ie8还是不行
3、eval中使用function xx(){}不能定义函数，需要加上window['xx'] = xx;
'''

import time
import logging
import win32gui, win32con
import win32com.client.dynamic
import pythoncom
import pywintypes
import six

SID_SWebBrowserApp = pywintypes.IID('{0002DF05-0000-0000-C000-000000000046}')

class IEDriverError(RuntimeError):
    '''
    '''

class IEDriver(object):
    '''window['qt4w_driver_lib']
    '''

    def __init__(self, ie_server_hwnd):
        self._hwnd = ie_server_hwnd
        self._init_com_obj()
        
    def _init_com_obj(self):
        '''初始化com对象
        '''
        if hasattr(self, '_doc'): logging.debug('[IEDriver] re_init com_obj')
        # else: time.sleep(2)  # 部分IE10上发现打开页面时不sleep会导致拒绝访问错误
        msg = win32gui.RegisterWindowMessage('WM_HTML_GETOBJECT')
        for _ in range(3):
            try:
                ret, result = win32gui.SendMessageTimeout(self._hwnd, msg, 0, 0, win32con.SMTO_ABORTIFHUNG, 2000)
                ob = pythoncom.ObjectFromLresult(result, pythoncom.IID_IDispatch, 0)
                self._doc = win32com.client.dynamic.Dispatch(ob)
                self._win = self._doc.parentWindow
                self.handle_error_page(self._doc)
                break
            except pythoncom.com_error as e:
                if (e.args[0] % 0x100000000) == 0x800401F0:
                    # 尚未调用 CoInitialize。
                    pythoncom.CoInitialize()
                    return self._init_com_obj()
            except AttributeError as e:
                # 页面跳转时易发生此问题
                logging.debug(str(e))
                time.sleep(0.5)
        else:
            raise RuntimeError('初始化COM对象失败')
        
    def _retry_for_access_denied(self, func):
        '''IE中经常出现可重试解决的80070005错误
        '''
        timeout = 2
        time0 = time.time()
        err_msg = '拒绝访问'
        while time.time() - time0 < timeout:
            try:
                return func()
            except pywintypes.com_error as e:
                if (e.args[0] % 0x100000000) == 0x80020009:
                    err_msg = e.args[2][2].encode('utf8').strip()
                    logging.info('[IEDriver] retry error: %s' % err_msg)
                    if e.args[2][5] % 0x100000000 == 0x80070005:
                        # 拒绝访问,重试可以解决
                        time.sleep(0.05)
                        continue
                    elif e.args[2][5] % 0x100000000 == 0x80020101:
                        # 由于出现错误 80020101 而导致此项操作无法完成
                        # 一般是页面加载未完成
                        time.sleep(1)
                        timeout = 20
                        continue
                raise e
        else:
            raise IEDriverError(err_msg)
        
    def _check_valid(self):
        '''检查com对象的有效性
        '''
        try:
            self._doc._oleobj_.GetIDsOfNames('readyState')
            return True
        except pywintypes.com_error as e:
            if (e.args[0] % 0x100000000) == 0x80070005:
                self._init_com_obj()  # 重新初始化
                return False
            elif (e.args[0] % 0x100000000) == 0x800401F0:
                # 尚未调用 CoInitialize。
                pythoncom.CoInitialize()
                return self._check_valid()
            raise e
        
    def _get_document(self, frame):
        '''获取frame的IHTMLDocument2指针，此方法可以跨域
        '''
        try:
            return frame.document
        except pywintypes.com_error:
            iServiceProvider = frame._oleobj_.QueryInterface(pythoncom.IID_IServiceProvider)
            iWebBrowser2 = win32com.client.dynamic.Dispatch(iServiceProvider.QueryService(SID_SWebBrowserApp, pythoncom.IID_IDispatch))
            doc = iWebBrowser2.Document
            if not isinstance(doc, win32com.client.dynamic.CDispatch):
                # ie10以上无法跨域
                win = iWebBrowser2._oleobj_.Invoke(1034, 0, pythoncom.DISPATCH_PROPERTYGET, True)  # IHTMLDocument2 HRESULT parentWindow([out, retval] IHTMLWindow2** p);
                win = win32com.client.dynamic.Dispatch(win)
                doc = self._retry_for_access_denied(lambda: win.document)
            return doc

    def get_frames(self, doc):
        '''
        '''
        try:
            return doc.frames
        except pywintypes.com_error as e:
            if (e.args[0] % 0x100000000) == 0x80020009 and (e.args[2][5] % 0x100000000) == 0x80020003:
                return doc.GetElementsByTagName('iframe')
            else: raise

    def get_frame_window(self, win, frame_id, url):
        '''获取doc中id或name为frame_id，或者url匹配的frame的IHTMLWindow对象
        '''
        if not win: doc = self._doc
        else: doc = win.document
        frames = self.get_frames(doc)
        if isinstance(frame_id, six.string_types) and len(frame_id) > 0:
            try:
                frame = frames.item(frame_id)
                doc = self._get_document(frame)
                return doc.parentWindow
            except pywintypes.com_error as e:
                if (e.args[0] % 0x100000000) == 0x80020009 and (e.args[2][5] % 0x100000000) == 0x80020003:
                    # 找不到成员
                    raise RuntimeError('未找到id=%s 的frame' % frame_id)
                else: raise e
        else:
            # 使用url查找
            if url[0] == '/':
                # add domain
                parent_url = doc.url
                pos = parent_url.find('/', 8)
                url = parent_url[:pos] + url
            for i in range(frames.length):
                frame = frames.item(i)
                doc = self._get_document(frame)
                if doc.url == url: return doc.parentWindow
            else:
                raise RuntimeError('未找到url=%s 的frame' % url)

    def handle_error_page(self, doc):
        '''处理错误页面
        '''
        if doc.URL.startswith('res://ieframe.dll/invalidcert.htm'):
            # 证书错误页面
            target_url = None
            pos = doc.URL.find('#')
            if pos > 0:
                target_url = doc.URL[pos + 1:]
                
            for elem in doc.all:
                if elem.id == 'overridelink':
                    logging.info('[IEDriver] Auto click continue button, target url is %s' % target_url)
                    elem.click()
                    break
            else:
                raise RuntimeError('Find continue button failed')

            if target_url:
                time0 = time.time()
                timeout = 30
                while time.time() - time0 < timeout:
                    if not doc.URL.startswith('res://') and doc.readyState == 'complete':
                        break
                    time.sleep(0.5)
                else:
                    logging.warn('Waiting for jump to target url timeout')
                    
                    
    def eval_script(self, frame_win, script, use_eval=True):
        '''
        IE10以上异常对象才有stack属性
        '''
        # script = 'document.script_result = function(){try{%s}catch(){}}();' % script #window.script_result无法获取
        logging.debug('[IEDriver] eval script: %s' % script[:200].strip())
        if not isinstance(script, six.text_type):
            script = script.decode('utf8')  # 必须使用unicode编码

        if use_eval:
            script = script.replace('\\', r'\\')
            script = script.replace('"', r'\"')
            script = script.replace('\r', r'\r')
            script = script.replace('\n', r'\n')
            script = r'''document.script_result = (function(){
                try{
                    var result = eval("%s");
                    if(result != undefined){
                        return 'S'+result.toString();
                    }else{
                        return 'Sundefined';
                    }
                }catch(e){
                    var retVal = 'E['+e.name + ']' + e.message;//toString()
                    if(e.stack) retVal += '\n' + e.stack;
                    else{
                        var f = arguments.callee.caller;
                        while (f) {
                            retVal += f.name;
                            f = f.caller;
                        }
                    }
                    return retVal;
                }
            })();''' % script  #
        # print script[:500]
        self._check_valid()
        
        if frame_win == None:
            frame_win = self._win
            frame_doc = self._doc
        else:
            frame_doc = frame_win.document

        self._retry_for_access_denied(lambda: frame_win.execScript(script)) # 
        if not use_eval: return
        
        if not self._check_valid(): return  # 一般是页面发生跳转，此时无法获取到直接结果
        
        name_id = frame_doc._oleobj_.GetIDsOfNames('script_result')
        result = frame_doc._oleobj_.Invoke(name_id, 0, pythoncom.DISPATCH_PROPERTYGET, True)
        if result == '': raise IEDriverError('JavaScript返回为空')
        if six.PY2 and isinstance(result, unicode):
            result = result.encode('utf8')
        logging.debug('[IEDriver] result: %s' % result[:200].strip())
        return result
    
if __name__ == '__main__':
    pass 



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#

'''文件窗口模块
'''

from qt4c.qpath import QPath
from qt4c.control import ControlContainer
import qt4c.wincontrols as win
import six

class FileDialog(win.Window, ControlContainer):
    '''文件窗口基类
    '''
    def __init__(self,qpath=None):
        '''constructor
        '''
        win.Window.__init__(self, locator=qpath)
        ControlContainer.__init__(self)
        locators = {
#            '打开按钮'     :{'type':win.Control, 'root':self, 'locator': QPath('/Caption = "打开(&O)" && visible="True"')}, #打开按钮
            '文件名编辑框' :{'type':win.Control, 'root':self, 'locator': QPath('/ClassName="ComboBoxEx32" && Visible="True"')},#顶上那个combobox
            '文件.文件路径展示框' :{'type':win.ComboBox, 'root':self, 'locator': QPath('/ControlId = "0x471" && visible="True" && MaxDepth = "2"')},#显示当前文件夹名称的combox
             '取消按钮':{'type':win.Control, 'root':self, 'locator': QPath('/Caption ~= "取消" && visible="True"')}, #取消按钮
                    }
        self.updateLocator(locators)
    
    @property
    def FilePath(self):
        '''返回当前文件路径。如果没有选择文件，返回的是当前文件夹路径
        '''
        import ctypes
        import win32con, win32gui
        import qt4c.util as util
        
        hwnd = self.HWnd
        size = win32con.MAX_PATH
        pid = self.ProcessId
        pm = util.ProcessMem(pid, size)
        msgid = win32con.CDM_GETFILEPATH
        rdsize = win32gui.SendMessage(hwnd, msgid, size, pm.Buffer)
        if rdsize <= 0:
            return None
        else:
            buff = ctypes.create_unicode_buffer(rdsize)
            pm.read(buff, ctypes.sizeof(buff))
            if six.PY2:
                return buff.value.encode('utf8')
            return buff.value
        
class OpenFileDialog(FileDialog):
    '''打开文件窗口
    '''    
    def __init__(self):
        FileDialog.__init__(self,QPath("/Caption ~= '打开' && Classname ~= '#32770' && Visible='True'"))
        locators = {
            '打开按钮'   :{'type':win.Control, 'root':self, 'locator': QPath('/Caption = "打开(&O)" && visible="True"')}, #打开按钮
            '图像.文件类型'   :{'type':win.Control, 'root':self, 'locator': QPath('/Caption~= "图像文件" && visible="True"')}, #打开图片窗口的文件类型Combox
            '文件.文件类型':{'type':win.Control, 'root':self, 'locator': QPath('/Caption~= "All Files" && visible="True"')}, #打开文件窗口的文件类型Combox
            '输入框':{'type':win.Control, 'root':self, 'locator': QPath("/ClassName='ComboBoxEx32'&&Visible='True'/ClassName='ComboBox'/ClassName='Edit'")}, #输入框              
            '内容显示区':{'type':win.Control, 'root':self, 'locator':QPath("/Caption='FolderView'&&Visible='True'&&MaxDepth = '2'") }, #输入框    
             '打开.文件浏览窗口':{'type':win.Control, 'root':self, 'locator': QPath("/ClassName='SHELLDLL_DefView'&&Visible='True'")}, #XP打开文件里的文件浏览窗口      
              '打开.win7文件浏览窗口':{'type':win.Control, 'root':self, 'locator': QPath("/ClassName='SysListView32'&& Caption = 'FolderView' && Visible='True' && MaxDepth = '3'")}, #win7打开文件里的文件浏览窗口       
              }
        self.updateLocator(locators)
    
    def open(self,filename):
        
        self.bringForeground()
        self.Controls['文件名编辑框'].Text=filename
        self.Controls['打开按钮'].click()
        self.waitForInvalid(timeout=30)

class SelectFileDialog(FileDialog):
    '''选择文件/文件夹窗口，适用于1.90之后发送文件打开的窗口
    '''    

    def __init__(self):
        FileDialog.__init__(self,QPath("|Caption ~= '选择文件/文件夹' && Classname ~= '#32770' && Visible='True'"))
        locators = {
            '发送按钮'   :{'type':win.Control, 'root':self, 'locator': QPath('/Caption = "发送(&S)" && visible="True"')}, #打开按钮
            '图像.文件类型'   :{'type':win.Control, 'root':self, 'locator': QPath('/Caption~= "图像文件" && visible="True"')}, #打开图片窗口的文件类型Combox
            '文件.文件类型':{'type':win.Control, 'root':self, 'locator': QPath('/Caption~= "All Files" && visible="True"')}, #打开文件窗口的文件类型Combox
            '输入框':{'type':win.Control, 'root':self, 'locator': QPath("/ClassName='ComboBoxEx32'&&Visible='True'/ClassName='ComboBox'/ClassName='Edit'")}, #输入框              
            'xp内容显示区':{'type':win.Control, 'root':self, 'locator':QPath("/Caption='FolderView'&&Visible='True'&&MaxDepth = '2'") }, #输入框
            'win7内容显示区':{'type':win.Control, 'root':self, 'locator':QPath("/ClassName='DirectUIHWND'&&Visible='True'&&MaxDepth = '5'") }, #输入框    
            '打开.文件浏览窗口':{'type':win.Control, 'root':self, 'locator': QPath("/ClassName='SHELLDLL_DefView'&&Visible='True'&&MaxDepth = '4'")}, #XP打开文件里的文件浏览窗口      
            '打开.win7文件浏览窗口':{'type':win.Control, 'root':self, 'locator': QPath("/ClassName='SHELLDLL_DefView'&&Visible='True'&&MaxDepth = '4'")}, #win7打开文件里的文件浏览窗口       
              }
        self.updateLocator(locators)
    
    def open(self,filename):
        self.bringForeground()
        self.Controls['文件名编辑框'].Text = filename
        self.Controls['发送按钮'].click()
        self.waitForInvalid(timeout=30)
        
class SaveAsDialog(FileDialog):
    '''另存为窗口
    '''    
    def __init__(self):
        FileDialog.__init__(self,QPath("/Caption ~= '存' && Classname = '#32770' && Visible='True'"))
        locators = {
             '保存按钮'   :{'type':win.Control, 'root':self, 'locator': QPath('/Caption = "保存(&S)" && visible="True"')}, #打开按钮
            '取消按钮'   :{'type':win.Control, 'root':self, 'locator': QPath('/Caption = "取消" && visible="True"')}, #打开按钮
            '输入框'    :{'type':win.Control, 'root':self, 'locator': QPath('/ClassName="ComboBoxEx32" && Visible="True"/ClassName="ComboBox" /ClassName="Edit"')},#xp
            'win7输入框'     :{'type':win.Control, 'root':self, 'locator': QPath('/ClassName="Edit"&&MaxDepth = "5"')},#win7
            '保存类型'   :{'type':win.ComboBox, 'root':self, 'locator': QPath('/Caption ~= "\." &&ClassName="ComboBox" && Visible="True" ')}, #保存类型Combox            
            '保存在'   :{'type':win.Control, 'root':self, 'locator':  QPath('/ClassName="ComboBox" && Visible="True" &&Instance = "0"')}, #xp保存类型Combox            
            '文件路径选择框'   :{'type':win.Control, 'root':self, 'locator':  QPath('/ClassName="Breadcrumb Parent" && Visible="True" &&MaxDepth = "5"/ClassName="ToolbarWindow32"')}, #win7上的文件路径
            '文件路径编辑框'   :{'type':win.Control, 'root':self, 'locator':  QPath('/ClassName="WorkerW"/ClassName="Edit"&&MaxDepth = "6"')}, #win7上的文件路径
                  }
        self.updateLocator(locators)
        
    def save(self, filepath, style=None):
        '''保存至路径
        
        :param filepath: 要保存至的全路径
        :type filepath: string
        :param style: 保存类型
        :type style: string
        '''
        self.bringForeground()
        import platform
        if platform.release() == 'XP':
            self.Controls['输入框'].Text = filepath
        else:
            self.Controls['win7输入框'].Text = filepath
        if style != None:
            self.Controls['保存类型'].SelectedIndex = style
        self.Controls['保存按钮'].click()
        self.waitForInvalid()
 

        
class BrowseDialog(FileDialog):
    '''浏览文件夹窗口
    '''    
    def __init__(self):
        FileDialog.__init__(self,QPath("/Caption = '浏览文件夹' && Classname = '#32770' && Visible='True'"))
        locators = {
            '保存按钮'   :{'type':win.ComboBox, 'root':self, 'locator': QPath('/Caption = "保存(&S)" && visible="True"')}, #打开按钮
            '保存类型'   :{'type':win.Control, 'root':self, 'locator': QPath('/Caption~="图像文件" && ClassName="ComboBox" && Visible="True"')}, #保存类型Combox            
             
                    }
        self.updateLocator(locators)
 

class BrowseFolderDialog(win.Window):
    '''浏览文件夹窗口
    '''
    def __init__(self):
        win.Window.__init__(self, locator=QPath("/Caption = '浏览文件夹' && Classname = '#32770' && Visible='True'"))
        locators = {
            '确定按钮'  :{'type':win.Control, 'root':self, 'locator': QPath("/Caption = '确定' && Visible='True' && maxDepth= '2'")},
            '取消按钮'  :{'type':win.Control, 'root':self, 'locator': QPath("/Caption = '取消' && Visible='True' && maxDepth= '2'")},
            '选择您要发送的文件夹':   {'type':win.TreeView, 'root':self, 'locator': QPath("/classname='SHBrowseForFolder ShellNameSpace Control' /ClassName='SysTreeView32' && Visible='True'")},           
                    }
        self.updateLocator(locators)
        
        
class FileFolder(FileDialog):
    '''打开文件窗口
    '''
    def __init__(self):
        FileDialog.__init__(self,QPath("/Classname= 'CabinetWClass' && Visible='True'"))
        locators = {
            '文件.文件浏览窗口'   :{'type':win.Control, 'root':self, 'locator': QPath('/ControlId = "0x1" && visible="True" && MaxDepth="7"')}, #打开件夹后文件的展示窗口
            
              }
        self.updateLocator(locators)

class ExploreFileFolder(FileDialog):
    '''XP左边有文件树型结构的文件窗口
    '''
    def __init__(self):
        FileDialog.__init__(self,QPath("/Classname='ExploreWClass' && Visible='True'"))
        locators = {
            '文件地址显示框'   :{'type':win.Control, 'root':self, 'locator': QPath('/ControlId = "0xA205" && Classname="Edit" && visible="True" && MaxDepth="5"')}, 
            
              }
        self.updateLocator(locators)

class Win7ExploreFileFolder(FileDialog):
    '''win7下在aio中打开收到文件的文件夹窗口
    '''
    
    def __init__(self):
        FileDialog.__init__(self,QPath("/Classname='CabinetWClass' && Visible='True'"))
        locators = {
            '文件地址显示框'   :{'type':win.Control, 'root':self, 'locator': QPath('/ControlId = "0x3E9" && Classname="ToolbarWindow32" && visible="True" && MaxDepth="8"')}, 
            
              }
        self.updateLocator(locators)
        
if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
"""用于访问支持IAccessible接口的控件

"""

import win32com.client
import win32con
import win32gui
import ctypes
import comtypes
import six
from comtypes.client import GetModule
GetModule('oleacc.dll') 
from comtypes.gen.Accessibility import IAccessible
from comtypes.automation import VARIANT

class EnumAccessibleObjectRole(object):
    '''Accessible Object的角色
    '''
    ROLE_SYSTEM_TITLEBAR = 0x1
    ROLE_SYSTEM_MENUBAR = 0x2
    ROLE_SYSTEM_SCROLLBAR = 0x3
    ROLE_SYSTEM_GRIP = 0x4
    ROLE_SYSTEM_SOUND = 0x5
    ROLE_SYSTEM_CURSOR = 0x6
    ROLE_SYSTEM_CARET = 0x7
    ROLE_SYSTEM_ALERT = 0x8
    ROLE_SYSTEM_WINDOW = 0x9
    ROLE_SYSTEM_CLIENT = 0xa
    ROLE_SYSTEM_MENUPOPUP = 0xb
    ROLE_SYSTEM_MENUITEM = 0xc
    ROLE_SYSTEM_TOOLTIP = 0xd
    ROLE_SYSTEM_APPLICATION = 0xe
    ROLE_SYSTEM_DOCUMENT = 0xf
    ROLE_SYSTEM_PANE = 0x10
    ROLE_SYSTEM_CHART = 0x11
    ROLE_SYSTEM_DIALOG = 0x12
    ROLE_SYSTEM_BORDER = 0x13
    ROLE_SYSTEM_GROUPING = 0x14
    ROLE_SYSTEM_SEPARATOR = 0x15
    ROLE_SYSTEM_TOOLBAR = 0x16
    ROLE_SYSTEM_STATUSBAR = 0x17
    ROLE_SYSTEM_TABLE = 0x18
    ROLE_SYSTEM_COLUMNHEADER = 0x19
    ROLE_SYSTEM_ROWHEADER = 0x1a
    ROLE_SYSTEM_COLUMN = 0x1b
    ROLE_SYSTEM_ROW = 0x1c
    ROLE_SYSTEM_CELL = 0x1d
    ROLE_SYSTEM_LINK = 0x1e
    ROLE_SYSTEM_HELPBALLOON = 0x1f
    ROLE_SYSTEM_CHARACTER = 0x20
    ROLE_SYSTEM_LIST = 0x21
    ROLE_SYSTEM_LISTITEM = 0x22
    ROLE_SYSTEM_OUTLINE = 0x23
    ROLE_SYSTEM_OUTLINEITEM = 0x24
    ROLE_SYSTEM_PAGETAB = 0x25
    ROLE_SYSTEM_PROPERTYPAGE = 0x26
    ROLE_SYSTEM_INDICATOR = 0x27
    ROLE_SYSTEM_GRAPHIC = 0x28
    ROLE_SYSTEM_STATICTEXT = 0x29
    ROLE_SYSTEM_TEXT = 0x2a
    ROLE_SYSTEM_PUSHBUTTON = 0x2b
    ROLE_SYSTEM_CHECKBUTTON = 0x2c
    ROLE_SYSTEM_RADIOBUTTON = 0x2d
    ROLE_SYSTEM_COMBOBOX = 0x2e
    ROLE_SYSTEM_DROPLIST = 0x2f
    ROLE_SYSTEM_PROGRESSBAR = 0x30
    ROLE_SYSTEM_DIAL = 0x31
    ROLE_SYSTEM_HOTKEYFIELD = 0x32
    ROLE_SYSTEM_SLIDER = 0x33
    ROLE_SYSTEM_SPINBUTTON = 0x34
    ROLE_SYSTEM_DIAGRAM = 0x35
    ROLE_SYSTEM_ANIMATION = 0x36
    ROLE_SYSTEM_EQUATION = 0x37
    ROLE_SYSTEM_BUTTONDROPDOWN = 0x38
    ROLE_SYSTEM_BUTTONMENU = 0x39
    ROLE_SYSTEM_BUTTONDROPDOWNGRID = 0x3a
    ROLE_SYSTEM_WHITESPACE = 0x3b
    ROLE_SYSTEM_PAGETABLIST = 0x3c
    ROLE_SYSTEM_CLOCK = 0x3d
    ROLE_SYSTEM_SPLITBUTTON = 0x3e
    ROLE_SYSTEM_IPADDRESS = 0x3f
    ROLE_SYSTEM_OUTLINEBUTTON = 0x40

class EnumAccessibleObjectState(object):
    '''Accessible Object的状态
           
    '''
    STATE_SYSTEM_UNAVAILABLE = 0x00000001  # Disabled
    STATE_SYSTEM_SELECTED = 0x00000002
    STATE_SYSTEM_FOCUSED = 0x00000004
    STATE_SYSTEM_PRESSED = 0x00000008
    STATE_SYSTEM_CHECKED = 0x00000010
    STATE_SYSTEM_MIXED = 0x00000020  # 3-state checkbox or toolbar button
    STATE_SYSTEM_INDETERMINATE = STATE_SYSTEM_MIXED
    STATE_SYSTEM_READONLY = 0x00000040
    STATE_SYSTEM_HOTTRACKED = 0x00000080
    STATE_SYSTEM_DEFAULT = 0x00000100
    STATE_SYSTEM_EXPANDED = 0x00000200
    STATE_SYSTEM_COLLAPSED = 0x00000400
    STATE_SYSTEM_BUSY = 0x00000800
    STATE_SYSTEM_FLOATING = 0x00001000  # Children "owned" not "contained" by parent
    STATE_SYSTEM_MARQUEED = 0x00002000
    STATE_SYSTEM_ANIMATED = 0x00004000
    STATE_SYSTEM_INVISIBLE = 0x00008000
    STATE_SYSTEM_OFFSCREEN = 0x00010000
    STATE_SYSTEM_SIZEABLE = 0x00020000
    STATE_SYSTEM_MOVEABLE = 0x00040000
    STATE_SYSTEM_SELFVOICING = 0x00080000
    STATE_SYSTEM_FOCUSABLE = 0x00100000
    STATE_SYSTEM_SELECTABLE = 0x00200000
    STATE_SYSTEM_LINKED = 0x00400000
    STATE_SYSTEM_TRAVERSED = 0x00800000
    STATE_SYSTEM_MULTISELECTABLE = 0x01000000  # Supports multiple selection
    STATE_SYSTEM_EXTSELECTABLE = 0x02000000  # Supports extended selection
    STATE_SYSTEM_ALERT_LOW = 0x04000000  # This information is of low priority
    STATE_SYSTEM_HASSUBMENU = STATE_SYSTEM_ALERT_LOW
    STATE_SYSTEM_ALERT_MEDIUM = 0x08000000  # This information is of medium priority
    STATE_SYSTEM_ALERT_HIGH = 0x10000000  # This information is of high priority
    STATE_SYSTEM_PROTECTED = 0x20000000  # access to this is restricted
    STATE_SYSTEM_VALID = 0x3FFFFFFF
    STATE_SYSTEM_HASPOPUP = 0x40000000

class _AccessibleObjectWrapper_win32com(object):
    """使用win32com模块实现IAccessible接口的包裹类
    
    """
    def __init__(self, acc_disp):
        self._acc_disp = acc_disp
        self._childID = win32con.CHILDID_SELF

    @property
    def accChildCount(self):
        count = self._acc_disp._oleobj_.InvokeTypes(-5001, 0, 2, (3, 0), ())
        if isinstance(count, six.integer_types):
            return count
        return None

    @property
    def accFocus(self):
        child = self._acc_disp._oleobj_.InvokeTypes(-5011, 0, 2, (12, 0), ())
        if isinstance(child, six.integer_types):
            return child
        elif isinstance(child, win32com.client.DispatchBaseClass):
            return _AccessibleObjectWrapper_win32com(child)
        else:
            return None
        
    @property
    def accName(self):
        name = self._acc_disp._oleobj_.InvokeTypes(-5003, 0, 2, (8, 0), ((12, 17),), self._childID)
        if name:
            if six.PY2:
                return name.encode('utf8')
            return name
        return None

    def get_accName(self, childID):
        if not childID:
            childID = self._childID
        name = self._acc_disp._oleobj_.InvokeTypes(-5003, 0, 2, (8, 0), ((12, 17),), childID)
        if name:
            if six.PY2:
                return name.encode('utf8')
            return name
        return None
        
    @property
    def accRole(self):
        return self._acc_disp._oleobj_.InvokeTypes(-5006, 0, 2, (12, 0), ((12, 17),), self._childID)
        
    @property
    def accDescription(self):
        desc = self._acc_disp._oleobj_.InvokeTypes(-5005, 0, 2, (8, 0), ((12, 17),), self._childID)
        if desc:
            if six.PY2:
                return desc.encode('utf8')
            return desc
        return None
        
    @property
    def accState(self):
        return self._acc_disp._oleobj_.InvokeTypes(-5007, 0, 2, (12, 0), ((12, 17),), self._childID)
    
    @property
    def accValue(self):
        val = self._acc_disp._oleobj_.InvokeTypes(-5004, 0, 2, (8, 0), ((12, 17),), self._childID)
        if val:
            if six.PY2:
                return val.encode('utf8')
            return val
        return None
    
    @property
    def accParent(self):
        ret = self._acc_disp._oleobj_.InvokeTypes(-5000, 
                                                  0, 
                                                  2, 
                                                  (9, 0), 
                                                  ())
        if ret is not None:
            ret = _AccessibleObjectWrapper_win32com(win32com.client.Dispatch(ret, u'accParent', None))
        return ret
    
class _AccessibleObjectWrapper_comtypes(object):
    """使用comtypes模块实现IAccessible接口的包裹类
    
    """
    def __init__(self, acc_disp):
        """Constructor
        :type acc_disp: comtypes.gen.Accessibility.IAccessible or
                        int( for window handle) or
                        tuple((x,y), for location point )
        :param acc_disp: acc_disp指定类型的实例
        :raise: TypeError
        """
        if isinstance(acc_disp, int):
            self._acc_disp = self._accessible_object_from_window(acc_disp)
        elif isinstance(acc_disp, tuple):
            self._acc_disp = self._accessible_object_from_point(acc_disp)
        elif isinstance(acc_disp, IAccessible):
            self._acc_disp = acc_disp
        else:
            raise TypeError("acc_disp must be ctypes.POINTER(IAccessible) or int or tuple")
        
        self._childID = win32con.CHILDID_SELF

    def _accessible_object_from_window(self, hwnd):
        """返回句柄指定的AccessibleObject
        
        :type hwnd: int
        :param hwnd: 句柄
        :raises: ValueError
        :rtype: comtypes.gen.Accessibility.IAccessible
        """
        if not win32gui.IsWindow(hwnd):
            raise ValueError("window(%s) is not valid!" % hwnd)
        
        OBJID_WINDOW = 0
        OBJID_CLIENT = -4
        if (win32gui.GetWindowLong(hwnd, win32con.GWL_STYLE) & win32con.WS_CHILDWINDOW) > 0:
            objID = OBJID_CLIENT
        else:
            objID = OBJID_WINDOW
        accObj = ctypes.POINTER(IAccessible)()
        ctypes.oledll.oleacc.AccessibleObjectFromWindow(hwnd, 
                                                        objID,
                                                        ctypes.byref(IAccessible._iid_), 
                                                        ctypes.byref(accObj))
        return accObj
    
    def _accessible_object_from_point(self, pt):
        """返回坐标对应的AccessibleObject
        
        :type pt: tuple
        :param pt: (x,y)，相对于桌面的坐标
        :raises: ValueError
        :rtype: comtypes.gen.Accessibility.IAccessible
        """
        accObj = ctypes.POINTER(IAccessible)()
        varChild = VARIANT()
        
        x,y = pt
        hr = ctypes.oledll.oleacc.AccessibleObjectFromPoint( ctypes.wintypes.POINT(x,y),
                                                        ctypes.byref(accObj),
                                                        ctypes.byref(varChild))
        if hr != 0 or varChild.value is None:
            raise ValueError("Can not fetch Accessible Object from point(%s,%s)" % (x,y))
        
        if varChild.value == win32con.CHILDID_SELF:
            return accObj
        else:
            return accObj.accChild(varChild)

    @property
    def accChildCount(self):
        count = self._acc_disp.accChildCount
        if isinstance(count, six.integer_types):
            return count
        return None
        
    @property
    def accFocus(self):
        child = self._acc_disp.accFocus
        if isinstance(child, six.integer_types):
            return child
        elif isinstance(child, IAccessible):
            return _AccessibleObjectWrapper_comtypes(child.QueryInterface(IAccessible))
        else:
            return None
        
    @property
    def accName(self):
        try:
            name = self._acc_disp.accName(self._childID)
            if name:
                if six.PY2:
                    return name.encode('utf8')
                return name
            return None
        except comtypes.COMError:
            return None

    def get_accName(self, childID):
        if not childID:
            childID = self._childID
        try:
            name = self._acc_disp.accName(childID)
            if name:
                if six.PY2:
                    return name.encode('utf8')
                return name
            return None
        except comtypes.COMError:
            return None

        
    @property
    def accRole(self):
        try:
            return self._acc_disp.accRole(self._childID)
        except comtypes.COMError:
            return None
        
    @property
    def accDescription(self):
        try:
            desc = self._acc_disp.accDescription(self._childID)
            if desc:
                if six.PY2:
                    return desc.encode('utf8')
                return desc
            return None
        except comtypes.COMError:
            return None
        
    @property
    def accState(self):
        return self._acc_disp.accState(self._childID)
    
    @property
    def accValue(self):
        try:
            val = self._acc_disp.accValue(self._childID)
            if val:
                if six.PY2:
                    return val.encode('utf8')
                return val
            return None
        except comtypes.COMError:
            return None
        
    @property
    def accParent(self):
        ret = self._acc_disp.accParent
        if ret:
            return _AccessibleObjectWrapper_comtypes(ret.QueryInterface(IAccessible))
        return ret
    
class AccessibleObject(object):
    """支持IAccessible接口的对象（控件）的包裹类
    
    """
    #===========================================================================
    #win32控件和web控件的AccessibleObject都是使用了ctypes模块调用系统api函数得到的对象，
    #它的类型是ctypes.Struct。ctypes模块可以使用pythoncom(26).dll的导出函数PyObjectFromIUnknown
    #将IAccessible（ctypes.Struct类型）转换为PyIDispatch类型实例，但转换后的实例在调用IAccessible接口
    #方法时，在某些对象下（更多体现在web元素）时会抛出异常，而ctypes.Struct类型实例则正常调用。
    #故无法统一使用pythoncom方式，只能采用了两种模块来封装IAccessible接口。
    #===========================================================================
    def __init__(self, acc_disp):
        """ Constructor
        
        :type acc_disp: win32com.client.dynamic.CDispatch or
                        win32com.client.DispatchBaseClass or
                        comtypes.gen.Accessibility.IAccessible or
                        int( for window handle) or
                        tuple((x,y), for location point ) or
                        AccessibleObject
        :param acc_disp: acc_disp指定类型的实例
        :raise: TypeError
        """
        if isinstance(acc_disp, win32com.client.dynamic.CDispatch) or\
           isinstance(acc_disp, win32com.client.DispatchBaseClass):
            self._acc_disp = _AccessibleObjectWrapper_win32com(acc_disp)
        elif isinstance(acc_disp, ctypes.c_void_p) or\
             isinstance(acc_disp, int) or\
             isinstance(acc_disp, tuple) :
            self._acc_disp = _AccessibleObjectWrapper_comtypes(acc_disp)
        elif isinstance(acc_disp, _AccessibleObjectWrapper_win32com) or\
             isinstance(acc_disp, _AccessibleObjectWrapper_comtypes):
            self._acc_disp = acc_disp
        elif isinstance(acc_disp, AccessibleObject):
            self._acc_disp = acc_disp._acc_disp
        else:
            raise TypeError("参数必须是以下类型：win32com.client.dynamic.CDispatch or \
                                          win32com.client.DispatchBaseClass or\
                                          comtypes.gen.Accessibility.IAccessible or\
                                          int(window handle) or\
                                          tuple((x,y)) or\
                                          AccessibleObject, but real type is %s" % type(acc_disp))
        

    @property
    def accFocus(self):
        """获取具有焦点的控件
        
        :rtype: int or AccessibleObject or None
        :return: 如果返回为0代表具有焦点的控件是其本身，
                                             返回类型为整数，则代表其获得焦点的子控件的控件ID；
                                             返回类型为AccessibleObject，则代表其获得焦点的子控件实例；
                                             返回为None，代表未实现此接口。  
        """
        childFocus = self._acc_disp.accFocus
        if isinstance(childFocus, six.integer_types):
            return childFocus
        elif isinstance(childFocus, _AccessibleObjectWrapper_win32com) or\
             isinstance(childFocus, _AccessibleObjectWrapper_comtypes):
            return AccessibleObject(childFocus)
        else:
            return None 
        
    @property
    def accName(self):
        """获取名称
        
        :rtype: string
        """
        return self._acc_disp.accName
        
    @property
    def accRole(self):
        """获取角色
        
        :rtype: EnumAccessibleObjectRole
        """
        return self._acc_disp.accRole
        
    @property
    def accDescription(self):
        """获取描述
        
        :rtype: string
        """
        return self._acc_disp.accDescription
        
    @property
    def accState(self):
        """获取状态值
        
        :rtype: EnumAccessibleObjectState
        """
        return self._acc_disp.accState
    
    @property
    def accValue(self):
        """获取值
        
        :rtype: string
        """
        return self._acc_disp.accValue
    
    @property
    def accParent(self):
        """获取父控件
        
        :rtype: AccessibleObject
        """
        parent = self._acc_disp.accParent
        if parent:
            return AccessibleObject(parent)
        return parent

    @property
    def accChildCount(self):
        return self._acc_disp.accChildCount

    def get_accName(self, childID=None):
        return self._acc_disp.get_accName(childID)
            
            
if __name__ == '__main__':
    pass
    


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''
IE模块
'''

import logging
import os
import re
import subprocess
import time

import win32api
import win32com.client
import win32con
import win32event
import win32gui
import win32process

import qt4c.wincontrols as win32
from qt4c.app import App
from qt4c.qpath import QPath
from qt4c.util import Timeout
from qt4c.webview.iewebview import IEWebView
from qt4w.browser import IBrowser


class IEWindow(win32.Window):
    '''IE窗口 qt4w使用
    '''
    _timeout = Timeout(120, 1)

    def __init__(self, process_id):
        '''初始化，进程id

        :params process_id: 窗口进程id
        :type process_id: int
        '''
        self._pid = process_id
        qpstr = "|classname='IEFrame' && visible='True'|classname='WorkerW' && visible='True'\
        |classname='ReBarWindow32' |classname='Address Band Root' && Instance='0' |maxdepth='3' && classname='Edit' && ProcessId='%d'" % process_id
        old_timeout = win32.Control._timeout
        win32.Control._timeout = self._timeout
        addr_edit = win32.Control(locator=QPath(qpstr))
        win32gui.BringWindowToTop(addr_edit.TopLevelWindow.HWnd)  # 激活ie窗口，并显示在最前端
        win32.Control._timeout = old_timeout
        win32.Window.__init__(self, root=addr_edit.TopLevelWindow)
        # 实现窗口最大化的逻辑
        time.sleep(0.005)  # 优化最大化的视觉效果
#             win32gui.ShowWindow(addr_edit.TopLevelWindow.HWnd, win32con.SW_MAXIMIZE)

    @property
    def pid(self):
        return self._pid

    @property
    def ie_window(self):
        '''获取Internet Explorer_Server对应的ie窗口
        '''
        iever = IEBrowser.get_version()
        iever = int(iever.split('.')[0])
        if iever < 6:
            raise RuntimeError("不支持IE%s" % iever)
        elif iever == 6:  # ie6
            qp = QPath("/classname='Shell DocObject View' && visible='True'/classname='Internet Explorer_Server'")
        elif iever == 7:
            qp = QPath("/classname='TabWindowClass' && visible='True'/maxdepth='3' && classname='Internet Explorer_Server'")
        else:
            qp = QPath("/classname='Frame Tab' && visible='True'/maxdepth='3' && classname='Internet Explorer_Server'")

        ie_window = win32.Control(root=self, locator=qp)
        ie_window._timeout = self._timeout
        ie_window.HWnd
        return ie_window

    @property
    def webview(self):
        '''返回WebView

        :rtype: IEWebView
        :return: IEWebView，用于实例化对应的WebPage
        '''
        return IEWebView(self.ie_window)

    @property
    def Url(self):
        '''返回当前的URL地址
        '''
        qpstr = "/classname='WorkerW' && visible='True' /classname='ReBarWindow32' \
        /classname~='(Address|ComboBox)' && Instance='0' /maxdepth='3' && classname='Edit'"

        addr_edit = win32.Control(root=self, locator=QPath(qpstr))
        return addr_edit.Caption

class IEBrowser(IBrowser):
    '''IE浏览器
    '''

    def __init__(self):
        self._pid_list = []

    @staticmethod
    def get_version():
        """获取注册表中的IE版本
        """
        hkey = win32con.HKEY_LOCAL_MACHINE
        subkey = r'SOFTWARE\Microsoft\Internet Explorer'
        hkey = win32api.RegOpenKey(hkey, subkey)
        ver = win32api.RegQueryValueEx(hkey, 'Version')[0]
        win32api.RegCloseKey(hkey)
        return ver

    @staticmethod
    def get_path():
        '''获取注册表中IE安装位置
        '''
        hkey = win32con.HKEY_LOCAL_MACHINE
        subkey = r'SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\IEXPLORE.EXE'
        hkey = win32api.RegOpenKey(hkey, subkey)
        return win32api.RegEnumValue(hkey, 0)[1]

    @staticmethod
    def searh_ie_window(url):
        '''查找ie进程查找到就退出，现在无法解决url对应的标签不在IE最前面的问题
        '''
        wmi = win32com.client.GetObject('winmgmts:')
        for p in wmi.InstancesOf('win32_process'):
            # 找到第一个就退出
            if p.CommandLine and p.CommandLine.find('SCODEF') == -1 and not p.CommandLine.lower().find('iexplore.exe') == -1:
                ie_winow = IEWindow(p.ProcessId)
                # IE url中可能存在“/#/”这样的无用字段，固用“/”代替
                ie_url = ie_winow.Url.replace('%3A', ':').replace('%2F', '/').replace('%23', '#').replace('/#/', '/')
                if url == ie_url or re.match(url, ie_url):
                    return ie_winow
        else:
            raise RuntimeError('%s对应的ie窗口不存在' % url)

    @staticmethod
    def killall():
        '''kill掉所有IE进程
        '''
        wmi = win32com.client.GetObject('winmgmts:')
        for p in wmi.InstancesOf('win32_process'):
            if p.Name == 'iexplore.exe':
                try:
                    p.Terminate
                except:
                    pass

    def close(self):
        '''kill掉所有IE进程
        '''
        wmi = win32com.client.GetObject('winmgmts:')
        for p in wmi.InstancesOf('win32_process'):
            if p.ProcessId in self._pid_list:
                try:
                    p.Terminate
                except Exception as e:
                    logging.error('kill process failed, error: %s' % e)

    def open_url(self, url, page_cls=None):
        '''打开一个url，返回对应的webpage实例类

        :params url: url
        :type url: str
        :params page_cls: page实例类
        :type page_cls: class
        '''
        ie_path = IEBrowser.get_path()
        # -e可以实现以新进程的方式打开ie
        pid = win32process.CreateProcess(None, '%s -e %s ' % (ie_path, url), None, None, 0, win32con.CREATE_NEW_CONSOLE, None, None, win32process.STARTUPINFO())[2]
        if not pid in self._pid_list:
            self._pid_list.append(pid)
        handle = win32api.OpenProcess(win32con.PROCESS_QUERY_INFORMATION, False, pid)
        win32event.WaitForInputIdle(handle, 10000)
        return self._get_page_cls(pid, page_cls)

    def find_by_url(self, url, page_cls=None, timeout=10):
        '''通过url查找页面，支持正则匹配
        '''
        from qt4c.exceptions import TimeoutError
        time0 = time.time()
        while time.time() - time0 < timeout:
            try:
                ie_window = IEBrowser.searh_ie_window(url)
                if not ie_window.pid in self._pid_list:
                    self._pid_list.append(ie_window.pid)
                break
            except RuntimeError as e:
                logging.debug('[IEBrowser] search ie window failed: %s' % e)
                time.sleep(0.5)
        else:
            raise TimeoutError
        return self._get_page_cls(ie_window, page_cls)

    def _get_page_cls(self, process_id_or_window, page_cls=None):
        '''获取具体的webpage实例类
        '''
        if isinstance(process_id_or_window, int):
            webview = IEWindow(process_id_or_window).webview
        else:
            webview = process_id_or_window.webview
        if page_cls:
            return page_cls(webview)
        return webview


class IEApp(object):
    '''to be deleted
    '''

    @staticmethod
    def killAll():
        IEBrowser.killall()


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#

import os
from setuptools import setup, find_packages

BASE_DIR = os.path.realpath(os.path.dirname(__file__))
  
def generate_version():
    version = "1.0.0"
    if os.path.isfile(os.path.join(BASE_DIR, "version.txt")):
        with open("version.txt", "r") as fd:
            content = fd.read().strip()
            if content:
                version = content
    return version
  
def parse_requirements():
    reqs = []
    if os.path.isfile(os.path.join(BASE_DIR, "requirements.txt")):
        with open(os.path.join(BASE_DIR, "requirements.txt"), 'r') as fd:
            for line in fd.readlines():
                line = line.strip()
                if line:
                    reqs.append(line)
        return reqs

def get_description():
    with open(os.path.join(BASE_DIR, "README.md"), "rb") as fh:
        return fh.read().decode('utf8')


if __name__ == "__main__":
    setup(
      version=generate_version(),
      name="qt4c",
      packages=find_packages(exclude=("tests", "tests.*",)),
      include_package_data=True,
      package_data={'':['*.txt', '*.TXT'], },
      data_files=[(".", ["requirements.txt"])],
      description = "QTA test driver for Win32 application",
      long_description=get_description(),
      long_description_content_type="text/markdown",
      author="Tencent",
      license="Copyright(c)2010-2018 Tencent All Rights Reserved. ",
      install_requires=parse_requirements(),
      classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 2.7",
        "Operating System :: Microsoft :: Windows",
      ],
    )



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#

'''Chrome浏览器驱动
'''

import time
import json
import logging
import threading
import six
from testbase.util import smart_text
from six.moves import urllib

class ChromeDriverError(RuntimeError):
    '''Chrome驱动错误
    '''
    def __init__(self, code, msg):
        self._code = code
        self._msg = msg
    
    @property
    def code(self):
        return self._code
    
    @property
    def message(self):
        return self._msg
    
    def __str__(self):
        return '[%d] %s' % (self._code, self._msg)
    
class ChromeDriver(object):
    '''Chrome驱动
    '''
    inst_dict = {}  # 每个WebSocket地址只能创建一个实例
    
    def __init__(self, port):
        self._port = port
        
    def get_page_list(self):
        '''获取打开的页面列表
        '''
        result = []
        null_proxy_handler = urllib.request.ProxyHandler({})
        opener = urllib.request.build_opener(null_proxy_handler)
        urllib.request.install_opener(opener)  # 强制不使用代理
        req = urllib.request.Request(url='http://localhost:%d/json' % self._port,
                          headers={'User-Agent': 'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0)'})
        timeout = 10
        time0 = time.time()
        while time.time() - time0 < timeout:
            try:
                ret = urllib.request.urlopen(req).read()
                break
            except urllib.error.URLError as e:
                logging.warn('get_page_list error: %s' % e)
                time.sleep(1)
        else:
            raise RuntimeError('无法访问Chrome远程调试Server')

        page_list = json.loads(ret)
        for item in page_list:
            if item['type'] == 'page':
                page = {'url': item['url'], 'title': item['title']}
                if 'webSocketDebuggerUrl' in item:
                    page['websocketurl'] = item['webSocketDebuggerUrl']
                else:
                    page['websocketurl'] = 'ws://localhost:%s/devtools/page/%s' % (self._port, item['id'])
                result.append(page)
        return result
    
    def get_debugger(self, url=None, title=None):
        '''获取Web调试器
        '''
        if url is None and title is None:
            raise ValueError('url或title参数不能同时为空')
        page_list = self.get_page_list()
        ws_addr = None
        if len(page_list) == 1: 
            ws_addr = page_list[0]['websocketurl']
        else:
            for page in page_list:
                if (url and smart_text(page['url']) == smart_text(url) ) or (title and smart_text(page['title']) == smart_text(title) ):
                    ws_addr = page['websocketurl']
                    break
            else:
                raise RuntimeError('page %s not found' % (url or title))
        if not ws_addr in self.__class__.inst_dict:
            self.__class__.inst_dict[ws_addr] = WebkitDebugger(ws_addr)
        return self.__class__.inst_dict[ws_addr]
    
class WebkitDebugger(object):
    '''Webkit调试器
    '''
    def __init__(self, ws_addr):
        import websocket
        self._ws_addr = ws_addr
#         self._ws = websocket.WebSocket()
#         self._ws.connect(self._ws_addr)
        self._ws = websocket.WebSocketApp(self._ws_addr,
                              on_open=lambda ws: self.on_open(ws),
                              on_message=lambda ws, message: self.on_message(ws, message),
                              on_error=lambda ws, error: self.on_error(ws, error),
                              on_close=lambda ws: self.on_close(ws))
        self._seq = 0
        self._connected = False
        self._context_dict = {}
        self._data_dict = {}
        t = threading.Thread(target=self.work_thread)
        t.setDaemon(True)
        t.start()
        self._wait_for_ready()
        self._init()
    
    def __del__(self):
        self._ws.close()
        
    # ================= WebSocket回调 ===========================
    def on_open(self, ws):
        '''
        '''
        self._connected = True
        
    def on_message(self, ws, message):
        '''收到消息
        '''
        logging.debug('[ChromeDriver] recv %s' % message[:200])
        result = json.loads(message)
        if 'id' in result:
            self._data_dict[result['id']] = result
        else:
            self.on_recv_notify_msg(result['method'], result['params'])
            
    def on_error(self, ws, error):
        print(error)
     
    def on_close(self, ws):
        print("### closed ###")
    
    # ===========================================================
    
    def _wait_for_ready(self, timeout=10, interval=0.1):
        '''等待WebSocket连接
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self._connected: break
            time.sleep(interval)
        else:
            raise RuntimeError('Connect %s failed' % self._ws_addr)
    
    def _init(self):
        '''初始化
        '''
        self.enable_runtime()
    
    def _get_context_id(self, frame_id):
        '''获取contextId
        '''
        timeout = 5
        time0 = time.time()
        while time.time() - time0 < timeout:
            if frame_id in self._context_dict:
                time_cost = time.time() - time0
                if time_cost >= 0.1: logging.debug('[ChromeDriver] wait context id for %s cost %sS' % (frame_id, time_cost))
                return self._context_dict[frame_id]
            time.sleep(0.1)
        else:
            raise RuntimeError('Can\'t find Context Id of %s' % frame_id)
        
    def on_recv_notify_msg(self, method, params):
        '''接收到通知消息
        '''
        if method == 'Runtime.executionContextCreated':
            context = params['context']
            if 'type' in context and context['type'] == 'Extension': return
            if 'frameId' in context: frame_id = context['frameId']
            else: frame_id = context['auxData']['frameId']
            self._context_dict[frame_id] = context['id']
            logging.debug('[ChromeDriver] add context id: %s(%s %s)' % (context['id'], frame_id, context['origin']))
        elif method == 'Runtime.executionContextDestroyed':
            context_id = params['executionContextId']
            for frame in self._context_dict:
                if self._context_dict[frame] == context_id:
                    del self._context_dict[frame]
                    break
        else:
            pass
            
    def work_thread(self):
        '''工作线程
        '''
        self._ws.run_forever()
    
    def _wait_for_response(self, seq, timeout=600, interval=0.1):
        '''等待返回数据
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            if seq in self._data_dict: 
                result = self._data_dict.pop(seq)
                if 'result' in result:
                    return result['result']
                elif 'error' in result:
                    logging.warn('[ChromeDriver] response error %s' % result)
                    raise ChromeDriverError(result['error']['code'], result['error']['message'])
                
            time.sleep(interval)
        else:
            raise RuntimeError('Wait for [%s] response timeout' % seq)
        
    def send_request(self, method, **kwds):
        '''发送请求
        
        :param method: 命令字
        :type method:  string
        '''
        self._seq += 1
        req = {'id': self._seq, 'method': method}
        if kwds: req['params'] = kwds
        # print req
        data = json.dumps(req)
        self._ws.send(data)
        # logging.debug('[ChromeDriver] send %s' % data[:500])
        return self._wait_for_response(self._seq)
    
    def enable_runtime(self):
        '''
        '''
        self.send_request('Runtime.enable')
        
    def get_frame_tree(self):
        '''获取frame树
        '''
        result = self.send_request('Page.getResourceTree')
        return result['frameTree']
    
    def eval_script(self, frame_id, script):
        '''执行JavaScript
        '''
        if frame_id == None:
            frame_id = self.get_frame_tree()['frame']['id']
        context_id = self._get_context_id(frame_id)
        logging.debug('[ChromeDriver][Eval:%s] %s' % (context_id, script[:200]))
        script = script.replace('\\', r'\\')
        script = script.replace('"', r'\"')
        script = script.replace('\r', r'\r')
        script = script.replace('\n', r'\n')
        script = r'''(function(){
            try{
                var result = eval("%s");
                if(result != undefined){
                    return 'S'+result.toString();
                }else{
                    return 'Sundefined';
                }
            }catch(e){
                var retVal = 'E['+e.name + ']' + e.message;//toString()
                retVal += '\n' + e.stack;
                return retVal;
            }
        })();''' % script
        params = {'objectGroup': 'console', 'includeCommandLineAPI': True, 'doNotPauseOnExceptionsAndMuteConsole': False, 'returnByValue': False, 'generatePreview': True}
        result = self.send_request('Runtime.evaluate', contextId=context_id, expression=script, **params)
        result = result['result']['value']
        logging.debug('[ChromeDriver][Retn] %s' % result)
        return result

    def screenshot(self):
        '''通过Chrome开发者协议获取page页面截图，
        '''
        import base64
        logging.debug('[ChromeDriver][captureScreenshot]')
        params = {'format': 'png'}
        result = self.send_request('Page.bringToFront')
        result = self.send_request('Page.captureScreenshot', **params)
        image = result['data']
        image = base64.b64decode(image)
        from PIL import Image
        picture = Image.open(six.BytesIO(image))
        logging.debug('[ChromeDriver][Retn] %s' % result)
        return picture
        
if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#

'''PC端WebView基类
'''

import ctypes
import sys

import win32con
import win32gui

from qt4c.filedialog import FileDialog
from qt4c.mouse import Mouse, MouseFlag, MouseClickType
from qt4c.qpath import QPath
from qt4w.webview.webview import IWebView


class WebViewBase(IWebView):
    '''PC端WebView基类
    '''
    def __init__(self, window, webdriver, offscreen_win=None):
        self._window = window
        self._offscreen_win = offscreen_win
        self._webdriver = webdriver
        self._parent_wnd = win32gui.GetParent(self._window.HWnd)
        # 如果获取的父窗口是空
        if self._parent_wnd == 0:
            self._parent_wnd = self._window.HWnd
        self._browser_type = 'not defined'
        
    def __getattr__(self, attr):
        '''转发给WebDriver实现
        '''
        return getattr(self._webdriver, attr)
        
    @property
    def browser_type(self):
        return self._browser_type
    
    @browser_type.setter
    def browser_type(self, type_name):
        self._browser_type = type_name
    
    def _handle_result(self, result, frame_xpaths):
        '''处理执行JavaScript的结果
        
        :param result: 要处理的数据
        :type  result: string
        :param frame_xpaths: 执行js所在frame的xpath
        :type  frame_xpaths: list
        '''
        from qt4w.util import JavaScriptError
        if result[0] == 'S': return result[1:]
        elif result[0] == 'E':
            raise JavaScriptError(frame_xpaths, result[1:])
        else:
            raise ValueError('执行JavaScript返回结果错误：%r' % result)
    
    def _handle_offset(self, x_offset, y_offset):
        '''win10上如果设置了DPI需要进行坐标修正
        '''
        from qt4c import util
        ratio = util.getDpi(self._window.HWnd)
        return x_offset / ratio, y_offset / ratio

    def _inner_click(self, flag, click_type, x_offset, y_offset,):
        self.activate()
        x_offset, y_offset = self._handle_offset(x_offset, y_offset)
        new_x, new_y = win32gui.ClientToScreen(self._window.HWnd, (int(x_offset), int(y_offset)))
        if self._offscreen_win:
            new_x += self._window.BoundingRect.Left - self._offscreen_win.BoundingRect.Left
            new_y += self._window.BoundingRect.Top - self._offscreen_win.BoundingRect.Top
        Mouse.click(new_x, new_y, flag, click_type)
        
    def click(self, x_offset, y_offset):  
        self._inner_click(MouseFlag.LeftButton, MouseClickType.SingleClick, x_offset, y_offset)
        
    def double_click(self, x_offset, y_offset):
        self._inner_click(MouseFlag.LeftButton, MouseClickType.DoubleClick, x_offset, y_offset)
    
    def right_click(self, x_offset, y_offset):
        self._inner_click(MouseFlag.RightButton, MouseClickType.SingleClick, x_offset, y_offset)
    
    def long_click(self, x_offset, y_offset, duration=1):
        raise NotImplementedError
        
    def hover(self, x_offset, y_offset):
        self.activate()
        new_x, new_y = win32gui.ClientToScreen(self._window.HWnd, (int(x_offset), int(y_offset)))
        if self._offscreen_win:
            new_x += self._window.BoundingRect.Left - self._offscreen_win.BoundingRect.Left
            new_y += self._window.BoundingRect.Top - self._offscreen_win.BoundingRect.Top
        Mouse.move(new_x, new_y)
    
    def scroll(self, backward=True):
        self._window.scroll(backward)
        
    def send_keys(self, keys):
        self._window.sendKeys(keys)
        
    def activate(self, is_true=True):
        '''激活当前窗口
        
        :param is_true: 是否激活，默认为True
        :type  is_true: bool
        '''
        # win32gui.SetWindowPos(self._parent_wnd,win32con.HWND_TOPMOST,0,0,0,0,win32con.SWP_NOSIZE|win32con.SWP_NOMOVE)
        self._window.TopLevelWindow.bringForeground()
        win32gui.BringWindowToTop(self._window.TopLevelWindow.HWnd)
        
    @property
    def rect(self):
        '''当前可见窗口的坐标信息
        '''
        return win32gui.GetWindowRect(self._window.HWnd)

    def screenshot(self):
        '''当前WebView的截图
        :return: PIL.Image
        '''
        self.activate()
        from PIL import ImageGrab
        from qt4c import util
        ratio = util.getDpi(self._window.HWnd)
        rect = tuple(map(lambda x: x * ratio, self.rect))
        img = ImageGrab.grab(rect)
        return img
    
    def upload_file(self, file_path):
        file_dialog = UploadFileDialog(self._window.ProcessId)
        file_dialog.upload_file(file_path)

class UploadFileDialog(FileDialog):
    '''上传文件对话框，目前ie、chrome封装是一样的，故放在这里，后面如果有不同，
    '''
    def __init__(self, process_id):
        '''通过进程id查找文件窗口
        '''
        qp = QPath('|classname~="32770" && ProcessId="%s"' % process_id)
        super(UploadFileDialog, self).__init__(qp)
        import qt4c.wincontrols as win32
        self.updateLocator({
            '文件名展示框' : { 'type':win32.Control, 'root':self,
                        'locator':QPath('/ClassName="ComboBoxEx32" && Visible="True" /ClassName="ComboBox" && Visible="True"') },
            '文件名' : { 'type':win32.Control, 'root':'@文件名展示框', 'locator':QPath('/ClassName="Edit" && Visible="True"') },
            
            '打开按钮' : { 'type':win32.Control, 'root':self, 'locator':QPath('/ClassName="Button" && Text="打开(&O)"') }
        })
        
    def upload_file(self, file_path):
        '''上传文件
        '''
        self.Controls['文件名'].Text = file_path
        self.Controls['打开按钮'].click()



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''
qpath模块

详见QPath类说明
'''

import re
import pythoncom
import win32gui
import winerror
from pkg_resources import iter_entry_points

from qt4c import wincontrols
from qt4c import uiacontrols
from qt4c import util
from qt4c.exceptions import ControlExpiredError,ControlAmbiguousError,ControlNotFoundError
import testbase.logger as logger
import six

class EnumQPathKey(object):
    MAX_DEPTH = "MAXDEPTH"
    INSTANCE = "INSTANCE"
    UI_TYPE = "UITYPE"

class EnumUIType(object):
    WIN = 'Win'
    UIA = 'UIA'

class QPathError(Exception):
    """QPath异常类定义
    """
    pass


class QPath(object):
    '''Query Path类，使用QPath字符串定位UI控件
    
    QPath的定义：
    Qpath ::= Seperator UIObjectLocator Qpath
    Seperator ::= 路径分隔符，任意的单个字符
    UIObjectLocator ::= UIObjectProperty && UIObjectLocator
    UIObjectProperty ::= UIProperty | RelationProperty | IndexProperty | UITypeProperty 
    UIProperty ::= [Window/UIA/Html]Property Operator “Value”
    UITypeProperty ::= Win | UIA | Html
    RelationProperty ::= MaxDepth = "Number"(最大搜索子孙深度)
    IndexProperty ::= Instance="Integer"(Integer:找到的多个控件中的第几个（负数表示从后往前数）)
    
    Operator ::= '=' | '~=' ('=' 表示精确匹配; '~=' 表示用正则表达式匹配) 
     
    UI控件基本上都是由树形结构组织起来的。为了方便定位树形结构的节点，QPath采用了路径结构
    的字符串形式。 QPath以第一个字符为路径分隔符，如 "/Node1/Node2/Node3"和 “|Node1|Node2|Node3"是一样的路径，都表示先找到Node1，再在Node1的子孙节点里找Node2，然后在Node2的子孙节点里
    找Node3。而定位每个Node需要改节点的多个属性以"&&"符号连接起来, 形成    "/Property1='value1' && property2~='value2' && ..."的形式，其中"~="表示正则匹配。QPath支持的属性包括wincontrols.Window和htmlcontrols.HtmlElement的类属性。QPath中的每个节点都有两个默认属性'UIType'和'MaxDepth'。 "UIType='Win32|UIA|Html'"，三个取值分别对应了三种QPath支持的UI类型。当UIType没有指定时默认取值为父节点的值。而"MaxDepth"表示该节点离祖先节点的最大深度，
    如果没有明确指定时默认取值为'1',即直接父子关系。QPath还支持"Instance”属性，用于当找到多个节点时指定选择第几个节点。
    
    例子：
    Qpath ="/ ClassName='TxGuiFoundation' && Caption~='QQ\d+' && Instance='-1' 
            / UIType='UIA' && name='mainpanel' && MaxDepth='10'"
    '''
    
    PROPERTY_SEP = '&&'
    OPERATORS = ["=", "~="]
    MATCH_FUNCS = {}
    MATCH_FUNCS["="]=lambda x, y: x == y
    MATCH_FUNCS["~="]= lambda string, pattern: re.search(pattern, string) != None
    CONTROL_TYPES = {
                   EnumUIType.WIN: wincontrols.Control,
                   EnumUIType.UIA: uiacontrols.Control
                   }
    
    def __init__(self, qpath_string):
        """Contructor
        
        :type qpath_string: string
        :param qpath_string: QPath字符串   
        """
        self.update_control_type()
        if not isinstance(qpath_string, six.string_types):
            raise QPathError("输入的QPath(%s)不是字符串!"  % (qpath_string) )
        self._strqpath = qpath_string
        self._path_sep, self._parsed_qpath = self._parse(qpath_string)
        self._error_qpath = None

    def update_control_type(self):
        for ep in iter_entry_points("qt4c.controls"):
            control_class = ep.load()
            self.CONTROL_TYPES[ep.name] = control_class
     

    def _find_controls_recur(self, root, qpath):
        '''递归查找控件
        
        :param root: 根控件
        :param qpath: 解析后的qpath结构
        :return: 返回(found_controls, remain_qpath)， 其中found_controls是找到的控件，remain_qpath
        是未能找到控件时剩下的未能匹配的qpath。
        '''
        qpath = qpath[:]
        props = qpath[0]
        props = dict((entry[0].upper(), entry[1]) for entry in props.items()) #使属性值大小写不敏感
        max_depth = 1 #默认depth是1
        if ( EnumQPathKey.MAX_DEPTH in props):
            max_depth = int(props[EnumQPathKey.MAX_DEPTH][1])
            if max_depth <= 0 :
                raise QPathError("MaxDepth=%s应该>=1" % max_depth)
            del props[EnumQPathKey.MAX_DEPTH]
        
        instance = None #默认没有index属性
        if (EnumQPathKey.INSTANCE in props):
            instance = int(props[EnumQPathKey.INSTANCE][1])
            del props[EnumQPathKey.INSTANCE]
            
        children = None
        if EnumQPathKey.UI_TYPE in props:
            uitype = props[EnumQPathKey.UI_TYPE][1]
            del props[EnumQPathKey.UI_TYPE]  
            child_ctrl_type = self.CONTROL_TYPES[uitype]
            if not isinstance(root, child_ctrl_type):
                try:
                    children = [child_ctrl_type(root)]
                except:
                    children = []
        if children is None:
            try:
                children = root.Children
            except ControlExpiredError:
                children = []
                
        found_child_controls = []
        for ctrl in children:
            if(self._match_control(ctrl, props)):
                found_child_controls.append(ctrl)
            
            if(max_depth > 1): 
                props_copy = props.copy()
                props_copy[EnumQPathKey.MAX_DEPTH] = ['=', str(max_depth - 1)]
                _controls, _ = self._find_controls_recur(ctrl, [props_copy])
                found_child_controls += _controls
        if not found_child_controls:
            return [], qpath
        
        if instance != None:
            try:
                found_child_controls = [found_child_controls[instance]]
            except IndexError:
                return [], qpath
        
        qpath.pop(0)
        
        if not qpath: #找到控件
            return found_child_controls, qpath
            
        else: #在子孙中继续寻找
            found_ctrls = []
            error_path = qpath
            for root in found_child_controls:
                ctrls, remain_qpath = self._find_controls_recur(root, qpath)
                found_ctrls += ctrls
                if len(remain_qpath) < len(error_path):
                    error_path = remain_qpath
            # remove same control
            cpy_found_ctrls = found_ctrls[:]
            found_ctrls = []
            for ctrl in cpy_found_ctrls:
                if ctrl not in found_ctrls:
                    found_ctrls.append(ctrl)
                    
            return found_ctrls, error_path
    

    def _match_control(self, control, props):
        """控件是否匹配给定的属性
        
        :param control: 控件
        :param props: 要匹配的控件属性字典，如{'classname':['=', 'window']}
        """
        attrs = dict((attr.upper(), attr) for attr in dir(control))
        for propname in props:
            if not propname in attrs:
                return False
            
            try: 
                act_prop_value = getattr(control, attrs[propname])
            except pythoncom.com_error as e: 
                return False
            except win32gui.error as e:
                if e.winerror == winerror.ERROR_INVALID_WINDOW_HANDLE: #无效窗口句柄
                    return False
                else:
                    raise e
            except ControlExpiredError as e:
                return False
                                   
            operator, exp_prop_value = props[propname]
            if act_prop_value is None:
                return False
            
            if isinstance(act_prop_value, bool):
                if exp_prop_value.upper() == 'TRUE':
                    exp_prop_value = True
                elif exp_prop_value.upper() == 'FALSE':
                    exp_prop_value = False
                else:
                    raise QPathError('不正确的bool属性值:%s' % exp_prop_value)
                if act_prop_value != exp_prop_value:
                    return False
            
            elif isinstance(act_prop_value, six.integer_types):
                if re.search('^0x', exp_prop_value) != None:
                    exp_prop_value = int(exp_prop_value, 16)
                else:
                    exp_prop_value = int(exp_prop_value)
                if act_prop_value != exp_prop_value:
                    return False
            
            elif isinstance(exp_prop_value, six.string_types):
                if not self.MATCH_FUNCS[operator](util.myEncode(act_prop_value), util.myEncode(exp_prop_value)):
                    return False
                    
            else:
                raise QPathError('不支持控件属性值类型：%s' % type(act_prop_value))
            
        return True            

    def _parse_property(self, prop_str):
        """解析property字符串，返回解析后结构
        
        例如将 "ClassName='Dialog' " 解析返回 {ClassName: ['=', 'Dialog']}
        """
        
        parsed_pattern = "(\w+)\s*([=~!<>]+)\s*[\"'](.*)[\"']"
        match_object = re.match(parsed_pattern, prop_str)
        if match_object is None:
            raise QPathError("属性(%s)不符合QPath语法" % prop_str)
        prop_name, operator, prop_value = match_object.groups()
        if not operator in self.OPERATORS:
            raise QPathError("QPath不支持操作符：%s"  % operator) 
        return {prop_name: [operator, prop_value]}
        
    def _parse(self, qpath_string):
        """解析qpath，并返回QPath的路径分隔符和解析后的结构
        
        将例如"| ClassName='Dialog' && Caption~='SaveAs' | UIType='UIA' && ControlID='123' && Instanc='-1'"
        的QPath解析为下面结构：[{'ClassName': ['=', 'Dialog'], 'Caption': ['~=', 'SaveAs']}, {'UIType': ['=', 'UIA'], 'ControlID': ['=', '123'], 'Instance': ['=', '-1']}]
        
        :param qpath_string: qpath 字符串
        :return: (seperator, parsed_qpath)
        """
        qpath_string = qpath_string.strip()
        seperator = qpath_string[0]
        locators = qpath_string[1:].split(seperator)
        
        parsed_qpath = []
        for locator in locators:
            props = locator.split(self.PROPERTY_SEP)
            parsed_locators = {}
            for prop_str in props:
                prop_str = prop_str.strip()
                if len(prop_str) == 0: 
                    raise QPathError("%s 中含有空的属性。" % locator)
                parsed_props = self._parse_property(prop_str)
                parsed_locators.update(parsed_props)
            parsed_qpath.append(parsed_locators)
        return seperator, parsed_qpath
       
    def __str__(self):
        '''返回格式化后的QPath字符串
        '''
        qpath_str = ""
        for locator in self._parsed_qpath:
            qpath_str += self._path_sep + " "
            delimit_str = " " + self.PROPERTY_SEP + " "
            locator_str = delimit_str.join(["%s %s '%s'"  % (key, locator[key][0], locator[key][1]) for key in locator])
            qpath_str += locator_str
        return qpath_str

    def getErrorPath(self):
        """返回最后一次QPath.search搜索未能匹配的路径
        
        :rtype: string
        """
        if self._error_qpath:
            props = self._error_qpath[0]
            delimit_str = " " + self.PROPERTY_SEP + " "
            return delimit_str.join(["%s %s '%s'" % (key, props[key][0], props[key][1]) for key in props]) 
        
    def search(self, root=None):
        """根据qpath和root查找控件
        
        :type root: 实例类型
        :param root:  查找开始的控件
        :return: 返回找到的控件列表
        """
        
        if root is None:
            root = wincontrols.Control() # desktop Control
        controls, self._error_qpath = self._find_controls_recur(root, self._parsed_qpath)
        return controls
    
def _find_by_name(root, name):
    qp = QPath("/Name='%s'&&MaxDepth='50'" % name)
    controls = qp.search(root)
    count = len(controls)
    if count > 1:
        logger.warning("根据qpath<%s>找到%d个控件，请优化qpath" % (qp._strqpath,count))
        return controls[0]
    elif count == 0:
        print(qp._strqpath)
        raise ControlNotFoundError("<%s>对应的控件不存在" % qp._strqpath)
    else:
        return controls[0]
        
if __name__ == '__main__':
    pass
